import copy
import os
from configparser import ConfigParser
from simses.commons.config.data.energy_management import EnergyManagementDataConfig
from simses.commons.config.data.power_electronics import PowerElectronicsConfig
from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.config.simulation.profile import ProfileConfig
from simses.commons.config.simulation.system import StorageSystemConfig
from simses.commons.data.data_handler import DataHandler
from simses.commons.log import Logger
from simses.commons.state.system import SystemState
from simses.logic.power_distribution.efficient import EfficientPowerDistributor
from simses.logic.power_distribution.equal import EqualPowerDistributor
from simses.logic.power_distribution.power_distributor import PowerDistributor
from simses.logic.power_distribution.soc import SocBasedPowerDistributor
from simses.logic.power_distribution.technology import TechnologyBasedPowerDistributor
from simses.system.auxiliary.heating_ventilation_air_conditioning.fix_cop_hvac import \
    FixCOPHeatingVentilationAirConditioning
from simses.system.auxiliary.heating_ventilation_air_conditioning.hvac import HeatingVentilationAirConditioning
from simses.system.auxiliary.heating_ventilation_air_conditioning.no_hvac import NoHeatingVentilationAirConditioning
from simses.system.dc_coupling.bus_charging_dc_coupling import BusChargingDcCoupling
from simses.system.dc_coupling.bus_charging_profile import BusChargingProfileDcCoupling
from simses.system.dc_coupling.dc_coupler import DcCoupling
from simses.system.dc_coupling.no_dc_coupling import NoDcCoupling
from simses.system.dc_coupling.usp_dc_coupling import USPDCCoupling
from simses.system.housing.abstract_housing import Housing
from simses.system.housing.forty_ft_container import FortyFtContainer
from simses.system.housing.no_housing import NoHousing
from simses.system.housing.twenty_ft_container import TwentyFtContainer
from simses.system.power_electronics.acdc_converter.abstract_acdc_converter import AcDcConverter
from simses.system.power_electronics.acdc_converter.bonfiglioli import BonfiglioliAcDcConverter
from simses.system.power_electronics.acdc_converter.fix_efficiency import FixEfficiencyAcDcConverter
from simses.system.power_electronics.acdc_converter.no_loss import NoLossAcDcConverter
from simses.system.power_electronics.acdc_converter.notton import NottonAcDcConverter
from simses.system.power_electronics.acdc_converter.notton_loss import NottonLossAcDcConverter
from simses.system.power_electronics.acdc_converter.sinamics import Sinamics120AcDcConverter
from simses.system.power_electronics.acdc_converter.stable import M2bAcDcConverter
from simses.system.power_electronics.acdc_converter.stacked import AcDcConverterIdenticalStacked
from simses.system.power_electronics.acdc_converter.sungrow import SungrowAcDcConverter
from simses.system.power_electronics.dcdc_converter.abstract_dcdc_converter import DcDcConverter
from simses.system.power_electronics.dcdc_converter.fix_efficiency import FixEfficiencyDcDcConverter
from simses.system.power_electronics.dcdc_converter.no_dcdc import NoDcDcConverter
from simses.system.power_electronics.dcdc_converter.no_loss import NoLossDcDcConverter
from simses.system.power_electronics.electronics import PowerElectronics
from simses.system.storage_system_ac import StorageSystemAC
from simses.system.storage_system_dc import StorageSystemDC
from simses.system.thermal.ambient.user_battery_temperature import UserBatteryTemperatureProfile
from simses.system.thermal.ambient.ambient_thermal_model import AmbientThermalModel
from simses.system.thermal.ambient.constant_temperature import ConstantAmbientTemperature
from simses.system.thermal.ambient.location_temperature import LocationAmbientTemperature
from simses.system.thermal.model.no_system_thermal_model import NoSystemThermalModel
from simses.system.thermal.model.system_thermal_model import SystemThermalModel
from simses.system.thermal.model.zero_d_room_thermal_model import ZeroDRoomThermalModel
from simses.system.thermal.model.zero_d_system_thermal_model import ZeroDSystemThermalModel
from simses.system.thermal.solar_irradiation.location import LocationSolarIrradiationModel
from simses.system.thermal.solar_irradiation.no_solar_irradiation import NoSolarIrradiationModel
from simses.system.thermal.solar_irradiation.solar_irradiation_model import SolarIrradiationModel
from simses.technology import hydrogen
from simses.technology import lithium_ion, redox_flow
from simses.technology.hydrogen.system import Hydrogen
from simses.technology.lithium_ion.battery import LithiumIonBattery
from simses.technology.lithium_ion.circuit import BatteryCircuit
from simses.technology.redox_flow.system import RedoxFlow
from simses.technology.storage import StorageTechnology
from simses.commons.config.simulation.energy_management import EnergyManagementConfig

class StorageSystemFactory:
    """
    The StorageSystemFactory instantiates all necessary and configured objects for AC and DC storage systems.
    """

    __lithium_ion_name: str = lithium_ion.__name__.split('.')[-1]
    __hydrogen_name: str = hydrogen.__name__.split('.')[-1]
    __redox_flow_name: str = redox_flow.__name__.split('.')[-1]

    def __init__(self, config: ConfigParser):
        self.__log: Logger = Logger(type(self).__name__)
        self.__simulation_config: ConfigParser = config
        self.__system_config: StorageSystemConfig = StorageSystemConfig(config)
        self.__general_config: GeneralSimulationConfig = GeneralSimulationConfig(config)
        self.__power_electronics_data_config: PowerElectronicsConfig = PowerElectronicsConfig()
        self.__profile_config: ProfileConfig = ProfileConfig(config)
        self.__config_battery: BatteryConfig = BatteryConfig(config)
        self.__energy_management_data_config = EnergyManagementDataConfig()
        self.__energy_management_config = EnergyManagementConfig(config)

    def create_acdc_converter(self, converter: str, max_power: float,
                              intermediate_cicuit_voltage: float) -> AcDcConverter:
        converter_configuration: dict = self.__system_config.acdc_converter
        converter_type = converter_configuration[converter][StorageSystemConfig.ACDC_CONVERTER_TYPE]
        acdc_converter: AcDcConverter = None
        if converter_type == NoLossAcDcConverter.__name__:
            self.__log.debug('Creating acdc converter as ' + converter_type)
            acdc_converter = NoLossAcDcConverter(max_power)
        elif converter_type == NottonAcDcConverter.__name__:
            self.__log.debug('Creating acdc converter as ' + converter_type)
            acdc_converter = NottonAcDcConverter(max_power)
        elif converter_type == FixEfficiencyAcDcConverter.__name__:
            self.__log.debug('Creating acdc converter as ' + converter_type)
            try:
                efficiency: float = float(converter_configuration[converter][StorageSystemConfig.ACDC_CONVERTER_EFFICIENCY])
            except (TypeError, IndexError):
                efficiency: float = 0.95
            acdc_converter = FixEfficiencyAcDcConverter(max_power, efficiency)
        elif converter_type == M2bAcDcConverter.__name__:
            self.__log.debug('Creating acdc converter as ' + converter_type)
            acdc_converter = M2bAcDcConverter(max_power)
        elif converter_type == NottonLossAcDcConverter.__name__:
            self.__log.debug('Creating acdc converter as ' + converter_type)
            acdc_converter = NottonLossAcDcConverter(max_power)
        elif converter_type == Sinamics120AcDcConverter.__name__:
            self.__log.debug('Creating acdc converter as ' + converter_type)
            acdc_converter = Sinamics120AcDcConverter(max_power, self.__power_electronics_data_config)
        elif converter_type == SungrowAcDcConverter.__name__:
            self.__log.debug('Creating acdc converter as ' + converter_type)
            acdc_converter = SungrowAcDcConverter(max_power)
        elif converter_type == BonfiglioliAcDcConverter.__name__:
            self.__log.debug('Creating acdc converter as ' + converter_type)
            acdc_converter = BonfiglioliAcDcConverter(max_power)
        else:
            options: [str] = list()
            options.append(NoLossAcDcConverter.__name__)
            options.append(NottonAcDcConverter.__name__)
            options.append(NottonLossAcDcConverter.__name__)
            options.append(FixEfficiencyAcDcConverter.__name__)
            options.append(Sinamics120AcDcConverter.__name__)
            options.append(M2bAcDcConverter.__name__)
            options.append(SungrowAcDcConverter.__name__)
            options.append(BonfiglioliAcDcConverter.__name__)
            raise Exception('ACDC converter ' + converter_type + ' is unknown. '
                                                                 'Following options are available: ' + str(options))
        return self.create_stacked_acdc_converter(converter, acdc_converter)

    def create_stacked_acdc_converter(self, converter: str, acdc_converter: AcDcConverter) -> AcDcConverter:
        try:
            converter_configuration: dict = self.__system_config.acdc_converter
            number_converters: int = int(converter_configuration[converter][StorageSystemConfig.ACDC_CONVERTER_NUMBERS])
            if number_converters > 1:
                try:
                    switch_value: float = float(
                        converter_configuration[converter][StorageSystemConfig.ACDC_CONVERTER_SWITCH])
                except IndexError:
                    switch_value: float = 1.0
                return AcDcConverterIdenticalStacked(number_converters, switch_value,
                                                     acdc_converter, self.__power_electronics_data_config)
        except (TypeError, IndexError):
            pass
        return acdc_converter

    def create_dcdc_converter(self, converter: str, intermediate_circuit_voltage: float) -> DcDcConverter:
        converter_configuration: dict = self.__system_config.dcdc_converter
        converter_type = converter_configuration[converter][StorageSystemConfig.DCDC_CONVERTER_TYPE]
        if converter_type == NoLossDcDcConverter.__name__:
            self.__log.debug('Creating dcdc converter as ' + converter_type)
            return NoLossDcDcConverter(intermediate_circuit_voltage)
        elif converter_type == NoDcDcConverter.__name__:
            self.__log.debug('Creating dcdc converter as ' + converter_type)
            return NoDcDcConverter(intermediate_circuit_voltage)
        elif converter_type == FixEfficiencyDcDcConverter.__name__:
            self.__log.debug('Creating dcdc converter as ' + converter_type)
            try:
                efficiency: float = float(converter_configuration[converter][StorageSystemConfig.DCDC_CONVERTER_EFFICIENCY])
                return FixEfficiencyDcDcConverter(intermediate_circuit_voltage, efficiency)
            except IndexError:
                return FixEfficiencyDcDcConverter(intermediate_circuit_voltage)
        else:
            options: [str] = list()
            options.append(NoLossDcDcConverter.__name__)
            raise Exception('DCDC converter ' + converter_type + ' is unknown. '
                                                                 'Following options are available: ' + str(options))

    def create_ambient_temperature_model(self) -> AmbientThermalModel:
        ambient_temperature_model: str = str(self.__system_config.ambient_temperature_model[StorageSystemConfig.AMBIENT_TEMPERATURE_TYPE])
        if self.__system_config.thermal_simulation:
            if ambient_temperature_model == ConstantAmbientTemperature.__name__:
                self.__log.debug('Creating ambient temperature model as ' + ambient_temperature_model)
                try:
                    constant_ambient_temperature: float = float(self.__system_config.ambient_temperature_model[StorageSystemConfig.AMBIENT_TEMPERATURE_CONSTANT])
                    return ConstantAmbientTemperature(constant_ambient_temperature)
                except IndexError:
                    self.__log.debug(
                        'Creating ambient temperature model as ' + ambient_temperature_model + 'with default value of constant ambient temperature = 25°C.')
                    return ConstantAmbientTemperature()
            elif ambient_temperature_model == LocationAmbientTemperature.__name__:
                self.__log.debug('Creating ambient temperature model as ' + ambient_temperature_model)
                return LocationAmbientTemperature(self.__profile_config, self.__general_config)
            else:
                options: [str] = list()
                options.append(ConstantAmbientTemperature.__name__)
                options.append(LocationAmbientTemperature.__name__)
                raise Exception('Ambient temperature model ' + ambient_temperature_model + ' is unknown. '
                                'Following options are available: ' + str(options))
        else:
            if ambient_temperature_model == UserBatteryTemperatureProfile.__name__:
                self.__log.debug('Creating ambient temperature model as ' + ambient_temperature_model)
                return UserBatteryTemperatureProfile(self.__profile_config, self.__general_config)
            else:
                self.__log.debug('Creating ambient temperature model as ' + ConstantAmbientTemperature.__name__ +
                                 ' with default value of constant ambient temperature = 25°C.')
                return ConstantAmbientTemperature()

    def create_solar_irradiation_model(self, housing: Housing) -> SolarIrradiationModel:
        if self.__system_config.thermal_simulation:
            solar_irradiation_model: str = str(
                self.__system_config.solar_irradiation_model[StorageSystemConfig.SOLAR_IRRADIATION_TYPE])
            if solar_irradiation_model == NoSolarIrradiationModel.__name__:
                self.__log.debug('Creating solar irradiation model as ' + solar_irradiation_model)
                return NoSolarIrradiationModel()
            elif solar_irradiation_model == LocationSolarIrradiationModel.__name__:
                if isinstance(housing, NoHousing):
                    raise Exception('Solar irradiation model ' + solar_irradiation_model + ' is incompatible with '
                                    + NoHousing.__name__ + '. Please select another housing model, or select '
                                    + NoSolarIrradiationModel.__name__ + '.')
                self.__log.debug('Creating solar irradiation model as ' + solar_irradiation_model)
                return LocationSolarIrradiationModel(self.__profile_config, self.__general_config, housing)
            else:
                options: [str] = list()
                options.append(NoSolarIrradiationModel.__name__)
                options.append(LocationSolarIrradiationModel.__name__)
                raise Exception('Solar irradiation model ' + solar_irradiation_model + ' is unknown. '
                                                                                       'Following options are available: ' + str(
                    options))
        else:
            self.__log.debug('Creating solar irradiation model as ' + NoSolarIrradiationModel.__name__)
            return NoSolarIrradiationModel()

    def create_thermal_model_from(self, hvac: HeatingVentilationAirConditioning,
                                  ambient_thermal_model: AmbientThermalModel, housing: Housing,
                                  dc_systems: [StorageSystemDC], ac_dc_converter: AcDcConverter,
                                  solar_irradiation_model: SolarIrradiationModel) -> SystemThermalModel:

        supported_housing = (TwentyFtContainer, FortyFtContainer)

        if self.__system_config.thermal_simulation:
            if isinstance(housing, NoHousing):
                if isinstance(hvac, NoHeatingVentilationAirConditioning):
                    self.__log.debug('Creating thermal model for cell type ' + ZeroDRoomThermalModel.__class__.__name__)
                    return ZeroDRoomThermalModel(ambient_thermal_model, self.__general_config, dc_systems, ac_dc_converter)
                else:
                    raise Exception('Housing type' + NoHousing.__name__ + ' is only compatible with HVAC type ' + NoHeatingVentilationAirConditioning.__name__ + '.')
            elif isinstance(housing, supported_housing):
                self.__log.debug('Creating thermal model for cell type ' + ZeroDSystemThermalModel.__class__.__name__)
                return ZeroDSystemThermalModel(ambient_thermal_model, housing, hvac, self.__general_config,
                                               dc_systems, ac_dc_converter, solar_irradiation_model)
            else:
                raise Exception('Check if selected housing model is currently supported for thermal simulation, '
                                'or contact the SimSES Development Team.')
        else:
            self.__log.debug('Creating thermal model for cell type ' + NoSystemThermalModel.__class__.__name__)
            return NoSystemThermalModel(ambient_thermal_model, self.__general_config, dc_systems)

    def create_housing_from(self, housing_name: str, ambient_thermal_model: AmbientThermalModel) -> Housing:
        if self.__system_config.thermal_simulation:
            housing_configuration: dict = self.__system_config.housing
            housing_type = housing_configuration[housing_name][StorageSystemConfig.HOUSING_TYPE]
            if housing_type == NoHousing.__name__:
                self.__log.debug('Creating Housing model ' + NoHousing.__class__.__name__)
                return NoHousing()
            elif housing_type == TwentyFtContainer.__name__:
                self.__log.debug('Creating Housing model ' + TwentyFtContainer.__class__.__name__)
                return TwentyFtContainer(housing_configuration[housing_name],
                                         ambient_thermal_model.get_initial_temperature())
            elif housing_type == FortyFtContainer.__name__:
                self.__log.debug('Creating Housing model ' + FortyFtContainer.__class__.__name__)
                return FortyFtContainer(housing_configuration[housing_name],
                                        ambient_thermal_model.get_initial_temperature())
            else:
                options: [str] = list()
                options.append(NoHousing.__name__)
                options.append(TwentyFtContainer.__name__)
                options.append(FortyFtContainer.__name__)
                raise Exception('Housing model ' + housing_type + ' is unknown. '
                                                                  'Following options are available: ' + str(options))
        else:
            self.__log.debug('Creating Housing model ' + NoHousing.__class__.__name__)
            return NoHousing()

    def create_hvac_from(self, hvac: str, housing: Housing) -> HeatingVentilationAirConditioning:
        if self.__system_config.thermal_simulation:
            hvac_configuration: dict = self.__system_config.hvac
            hvac_type = hvac_configuration[hvac][StorageSystemConfig.HVAC_TYPE]

            supported_housing = (TwentyFtContainer, FortyFtContainer)
            supported_housing_names = [model.__name__ for model in supported_housing]

            if hvac_type == NoHeatingVentilationAirConditioning.__name__:
                self.__log.debug('Creating HVAC model ' + NoHeatingVentilationAirConditioning.__class__.__name__)
                return NoHeatingVentilationAirConditioning()
            elif hvac_type == FixCOPHeatingVentilationAirConditioning.__name__:
                if not isinstance(housing, supported_housing):
                    raise Exception('HVAC model ' + hvac_type + ' is incompatible with ' + housing.__class__.__name__
                                    + '. Please select ' + NoHeatingVentilationAirConditioning.__name__
                                    + ', or select a supported housing model: ' + str(supported_housing_names))
                self.__log.debug('Creating HVAC model ' + FixCOPHeatingVentilationAirConditioning.__class__.__name__)
                return FixCOPHeatingVentilationAirConditioning(hvac_configuration[hvac])
            else:
                options: [str] = list()
                options.append(NoHeatingVentilationAirConditioning.__name__)
                options.append(FixCOPHeatingVentilationAirConditioning.__name__)
                raise Exception('HVAC model ' + hvac_type + ' is unknown. '
                                                            'Following options are available: ' + str(options))
        else:
            self.__log.debug('Creating HVAC model ' + NoHeatingVentilationAirConditioning.__class__.__name__)
            return NoHeatingVentilationAirConditioning()

    def create_system_state_from(self, system_id, storage_id) -> SystemState:
        state = SystemState(system_id, storage_id)
        state.set(SystemState.TIME, self.__general_config.start)
        state.set(SystemState.AC_FULFILLMENT, 1.0)
        return state

    def create_power_distributor_ac(self) -> PowerDistributor:
        power_distributor_type: str = self.__system_config.power_distributor_ac
        if power_distributor_type == EqualPowerDistributor.__name__:
            return EqualPowerDistributor()
        elif power_distributor_type == SocBasedPowerDistributor.__name__:
            return SocBasedPowerDistributor()
        else:
            options: [str] = list()
            options.append(EqualPowerDistributor.__name__)
            options.append(SocBasedPowerDistributor.__name__)
            raise Exception('AC power distributor ' + power_distributor_type + ' is unknown. '
                                                                               'Following options are available: ' + str(
                options))

    def create_power_distributor_dc(self, dc_systems: [StorageSystemDC],
                                    power_electronics: PowerElectronics) -> PowerDistributor:
        power_distributor_dc: [str] = self.__system_config.power_distributor_dc
        power_distributor_type: str = power_distributor_dc[StorageSystemConfig.DC_POWER_DISTRIBUTOR_TYPE]
        if power_distributor_type == EqualPowerDistributor.__name__:
            return EqualPowerDistributor()
        elif power_distributor_type == SocBasedPowerDistributor.__name__:
            return SocBasedPowerDistributor()
        elif power_distributor_type == EfficientPowerDistributor.__name__:
            return EfficientPowerDistributor(power_electronics.max_power)
        elif power_distributor_type == TechnologyBasedPowerDistributor.__name__:
            systems: dict = dict()
            for dc_system in dc_systems:
                system_id: int = dc_system.state.get(SystemState.SYSTEM_DC_ID)
                technology: str = type(dc_system.get_storage_technology()).__name__
                systems[system_id] = technology
            priorities: [str] = copy.deepcopy(power_distributor_dc)
            del priorities[StorageSystemConfig.DC_POWER_DISTRIBUTOR_TYPE]
            found: bool = True
            for system in systems.values():
                system_found = False
                for priority in priorities:
                    if priority == system:
                        system_found = True
                        break
                found &= system_found
            if not priorities or not found:
                options: [str] = list()
                options.append(LithiumIonBattery.__name__)
                options.append(RedoxFlow.__name__)
                options.append(Hydrogen.__name__)
                raise Exception('No priorities for technologies are given or not all technologies are considered. '
                                'Following options are available: ' + str(options))
            return TechnologyBasedPowerDistributor(priorities, systems)
        else:
            options: [str] = list()
            options.append(EqualPowerDistributor.__name__)
            options.append(SocBasedPowerDistributor.__name__)
            options.append(TechnologyBasedPowerDistributor.__name__)
            raise Exception('DC power distributor ' + power_distributor_type + ' is unknown. '
                                                                               'Following options are available: ' + str(
                options))

    def create_storage_systems_ac(self, data_export: DataHandler) -> [StorageSystemAC]:
        ambient_thermal_model: AmbientThermalModel = self.create_ambient_temperature_model()
        ac_systems: [[str]] = self.__system_config.storage_systems_ac
        res: [StorageSystemAC] = list()
        names: [str] = list()
        id_number: int = 0
        for system in ac_systems:
            id_number += 1
            name = system[StorageSystemConfig.AC_SYSTEM_NAME]
            if name in names:
                raise Exception('Storage system name ' + name + ' is not unique.')
            names.append(name)
            power: float = float(system[StorageSystemConfig.AC_SYSTEM_POWER])
            intermediate_circuit_voltage: float = float(system[StorageSystemConfig.AC_SYSTEM_DC_VOLTAGE])
            converter: str = system[StorageSystemConfig.AC_SYSTEM_CONVERTER]
            housing: str = system[StorageSystemConfig.AC_SYSTEM_HOUSING]
            hvac: str = system[StorageSystemConfig.AC_SYSTEM_HVAC]
            acdc_converter: AcDcConverter = self.create_acdc_converter(converter, power, intermediate_circuit_voltage)
            housing_model: Housing = self.create_housing_from(housing, ambient_thermal_model)
            solar_irradiation_model: SolarIrradiationModel = self.create_solar_irradiation_model(housing_model)
            heating_cooling: HeatingVentilationAirConditioning = self.create_hvac_from(hvac, housing_model)
            hvac_set_point: float = heating_cooling.get_set_point_temperature()
            power_electronics: PowerElectronics = PowerElectronics(acdc_converter)
            dc_systems: [StorageSystemDC] = self.create_storage_systems_dc(name, data_export, hvac_set_point, ambient_thermal_model,
                                                                           intermediate_circuit_voltage, id_number)
            system_thermal_model: SystemThermalModel = self.create_thermal_model_from(heating_cooling,
                                                                                      ambient_thermal_model,
                                                                                      housing_model, dc_systems,
                                                                                      acdc_converter,
                                                                                      solar_irradiation_model)
            power_distributor: PowerDistributor = self.create_power_distributor_dc(dc_systems, power_electronics)
            state = self.create_system_state_from(id_number, 0)
            dc_couplings: [DcCoupling] = self.create_dc_couplings(name)
            res.append(StorageSystemAC(state, data_export, system_thermal_model, power_electronics, dc_systems,
                                       dc_couplings, housing_model, power_distributor))
        return res

    def create_dc_couplings(self, name: str) -> [DcCoupling]:
        dc_systems: [[str]] = self.__system_config.storage_systems_dc
        res: [DcCoupling] = list()
        for dc_system in dc_systems:
            system = dc_system[0]
            dc_type: str = dc_system[1]
            if system == name and DcCoupling.__name__ in dc_type:
                if dc_type == NoDcCoupling.__name__:
                    res.append(NoDcCoupling())
                elif dc_type == BusChargingDcCoupling.__name__:
                    charging_power: float = float(dc_system[2])
                    generation_power: float = float(dc_system[3])
                    res.append(BusChargingDcCoupling(charging_power, generation_power))
                elif dc_type == BusChargingProfileDcCoupling.__name__:
                    capacity: float = float(dc_system[2])
                    file_name: str = dc_system[3]
                    file: str = os.path.join(self.__profile_config.technical_profile_dir, file_name)
                    res.append(BusChargingProfileDcCoupling(self.__general_config, capacity, file))
                elif dc_type == USPDCCoupling.__name__:
                    file_name: str = dc_system[2]
                    res.append(USPDCCoupling(self.__general_config, self.__profile_config, file_name))
                else:
                    options: [str] = list()
                    options.append(NoDcCoupling.__name__)
                    options.append(BusChargingDcCoupling.__name__)
                    raise Exception('DcCoupling ' + dc_type + ' is unknown. '
                                                              'Following options are available: ' + str(options))
        return res

    def create_storage_systems_dc(self, name: str, data_export: DataHandler,
                                  hvac_set_point: float, ambient_thermal_model: AmbientThermalModel,
                                  intermediate_circuit_voltage: float, system_id: int) -> [StorageSystemDC]:
        dc_systems: [[str]] = self.__system_config.storage_systems_dc
        res: [StorageSystemDC] = list()
        storage_id = 0
        for dc_system in dc_systems:
            try:
                system = dc_system[StorageSystemConfig.DC_SYSTEM_NAME]
                dc_type: str = dc_system[StorageSystemConfig.DC_SYSTEM_CONVERTER]
                converter_configuration: dict = self.__system_config.dcdc_converter
                dc_converter: str = converter_configuration[dc_type][StorageSystemConfig.DCDC_CONVERTER_TYPE]
                if system == name and DcDcConverter.__name__ in dc_converter:
                    storage_id += 1
                    dcdc_converter: DcDcConverter = self.create_dcdc_converter(dc_type, intermediate_circuit_voltage)
                    technologies: [str] = copy.deepcopy(dc_system)
                    del technologies[StorageSystemConfig.DC_SYSTEM_CONVERTER]
                    del technologies[StorageSystemConfig.DC_SYSTEM_NAME]
                    storage_technology: StorageTechnology = self.create_storage_technology(technologies, data_export,
                                                                                           hvac_set_point, ambient_thermal_model,
                                                                                           storage_id, system_id,
                                                                                           intermediate_circuit_voltage)
                    res.append(StorageSystemDC(system_id, storage_id, data_export, dcdc_converter, storage_technology))
                    storage_id += max(0, len(technologies) - 1)
            except IndexError as err:
                raise Exception(str(err) + ' could not be found')
            except KeyError as err:
                raise Exception(str(err) + ' could not be found')
        if not res:
            raise Exception('Storage system ' + name + ' has no storage data. Please specify your storage system.')
        return res

    def create_storage_technology(self, technologies: [str], data_export: DataHandler,
                                  hvac_set_point: float, ambient_thermal_model: AmbientThermalModel,
                                  storage_id: int, system_id: int, voltage: float) -> StorageTechnology:
        technology_configuration: dict = self.__system_config.storage_technologies
        storage_configuration: dict = dict()
        for technology in technologies:
            technology_set: list = technology_configuration[technology]
            storage_type: str = technology_set[StorageSystemConfig.STORAGE_TYPE]
            if storage_type not in storage_configuration.keys():
                storage_configuration[storage_type] = list()
            storage_configuration[storage_type].append(technology_set)
        if self.__lithium_ion_name in storage_configuration.keys():
            batteries: [LithiumIonBattery] = list()
            for technology_set in storage_configuration[self.__lithium_ion_name]:
                capacity: float = float(technology_set[StorageSystemConfig.STORAGE_CAPACITY])
                cell: str = technology_set[StorageSystemConfig.BATTERY_CELL]
                try:
                    soc: float = float(technology_set[StorageSystemConfig.BATTERY_SOC])
                except (IndexError, ValueError):
                    soc: float = self.__config_battery.soc
                try:
                    soh: float = float(technology_set[StorageSystemConfig.BATTERY_SOH])
                except (IndexError, ValueError):
                    soh: float = self.__config_battery.start_soh
                batteries.append(LithiumIonBattery(cell, voltage, capacity, soc, soh, data_export,
                                                   hvac_set_point, storage_id,
                                                   system_id, self.__simulation_config))
                storage_id += 1
            if len(batteries) > 1:
                return BatteryCircuit(batteries)
            else:
                return batteries[0]
        elif self.__redox_flow_name in storage_configuration.keys():
            for technology_set in storage_configuration[self.__redox_flow_name]:
                capacity: float = float(technology_set[StorageSystemConfig.STORAGE_CAPACITY])
                stack_type = technology_set[StorageSystemConfig.REDOX_FLOW_STACK]
                stack_module_power: float = float(technology_set[StorageSystemConfig.STACK_MODULE_POWER])
                try:
                    pump_algorithm = technology_set[StorageSystemConfig.REDOX_FLOW_PUMP_ALGORITHM]
                except IndexError:
                    pump_algorithm = 'Default'
                return RedoxFlow(stack_type, stack_module_power, voltage, capacity, pump_algorithm,
                                 hvac_set_point, data_export,
                                 storage_id, system_id, self.__simulation_config)
        elif self.__hydrogen_name in storage_configuration.keys():
            if type(ambient_thermal_model).__name__ not in [ConstantAmbientTemperature.__name__]:
                raise Exception('Hydrogen models work currently only with ' + ConstantAmbientTemperature.__name__ + '. '
                                'Please adapt your config.')
            for technology_set in storage_configuration[self.__hydrogen_name]:
                capacity: float = float(technology_set[StorageSystemConfig.STORAGE_CAPACITY])
                fuel_cell = technology_set[StorageSystemConfig.FUEL_CELL_TYPE]
                fuel_cell_nominal_power = float(technology_set[StorageSystemConfig.FUEL_CELL_POWER])
                electrolyzer = technology_set[StorageSystemConfig.ELECTROLYZER_TYPE]
                electrolyzer_nominal_power = float(technology_set[StorageSystemConfig.ELECTROLYZER_POWER])
                storage = technology_set[StorageSystemConfig.HYDROGEN_STORAGE]
                max_pressure = float(technology_set[StorageSystemConfig.HYDROGEN_TANK_PRESSURE])
                return Hydrogen(data_export, fuel_cell, fuel_cell_nominal_power, electrolyzer,
                                electrolyzer_nominal_power, storage, capacity, max_pressure,
                                ambient_thermal_model.get_initial_temperature(), system_id, storage_id,
                                self.__simulation_config)
        else:
            options: [str] = list()
            options.append(self.__lithium_ion_name)
            options.append(self.__redox_flow_name)
            options.append(self.__hydrogen_name)
            raise Exception('Storage technology ' + str(technologies) + ' is unknown. '
                                                                        'Following options are available: ' + str(
                options))

    def close(self):
        self.__log.close()



from configparser import ConfigParser

from simses.commons.config.data.data_config import DataConfig


class ElectrolyzerDataConfig(DataConfig):

    def __init__(self, path: str = None, config: ConfigParser = None):
        super().__init__(path, config)
        self.__section: str = 'ELECTROLYZER_DATA'

    @property
    def lookuptable_dir(self) -> str:
        """Returns directory of electrolyzer data files"""
        return self.get_data_path(self.get_property(self.__section, 'ELECTROLYZER_LOOKUPTABLE_DATA_DIR'))

    @property
    def parameters_dir(self) -> str:
        """Returns directory of electrolyzer data files"""
        return self.get_data_path(self.get_property(self.__section, 'ELECTROLYZER_PARAMETERS_DATA_DIR'))

    @property
    def pem_electrolyzer_pc_file(self) -> str:
        """Returns filename for polarisation curve for PEM-Electrolyzer"""
        return self.lookuptable_dir + self.get_property(self.__section, 'PEM_ELECTROLYZER_PC_FILE')

    @property
    def pem_electrolyzer_power_file(self) -> str:
        """Returns filename for power curve for PEM-Electrolyzer"""
        return self.lookuptable_dir + self.get_property(self.__section, 'PEM_ELECTROLYZER_POWER_FILE')

    @property
    def pem_electrolyzer_multi_dim_analytic_para_file(self) -> str:
        """Returns filename for power curve for PEM-Electrolyzer"""
        return self.parameters_dir + self.get_property(self.__section, 'PEM_ELECTROLYZER_MULTI_DIM_ANALYTIC_PARA_FILE')

    @property
    def alkaline_electrolyzer_multidim_lookup_currentdensity_file(self) -> str:
        """Returns filename for power curve for Alkaline Electrolyzer"""
        return self.lookuptable_dir + self.get_property(self.__section,'ALKALINE_ELECTROLYZER_CURRENT_FILE')

    @property
    def alkaline_electrolyzer_fit_para_file(self) -> str:
        """Returns filename for power curve for PEM-Electrolyzer"""
        return self.parameters_dir + self.get_property(self.__section, 'ALKALINE_ELECTROLYZER_FIT_PARA_FILE')



from configparser import ConfigParser

import numpy as np
import pandas
import pytest

from simses.analysis.data.energy_management import EnergyManagementData
from simses.analysis.data.system import SystemData
from simses.analysis.evaluation.economic.revenue_stream.energy_cost_reduction import EnergyCostReduction
from simses.commons.config.analysis.economic import EconomicAnalysisConfig
from simses.commons.config.simulation.general import GeneralSimulationConfig

# fixed input parameters
from simses.commons.state.energy_management import EnergyManagementState
from simses.commons.state.system import SystemState

years = 5
timestep = 3600
elec_price = 0.3
pv_feed = 0.1
load_const = 3000
pv_const = 2000


def create_general_config() -> GeneralSimulationConfig:
    simulation_config: ConfigParser = ConfigParser()
    simulation_config.add_section('GENERAL')
    simulation_config.set('GENERAL', 'TIME_STEP', str(timestep))
    return GeneralSimulationConfig(simulation_config)


def create_economic_analysis_config() -> EconomicAnalysisConfig:
    analysis_config: ConfigParser = ConfigParser()
    analysis_config.add_section('ECONOMIC_ANALYSIS')
    analysis_config.set('ECONOMIC_ANALYSIS', 'ELECTRICITY_PRICE', str(elec_price))
    analysis_config.set('ECONOMIC_ANALYSIS', 'PV_FEED_IN_TARIFF', str(pv_feed))
    return EconomicAnalysisConfig(analysis_config)


@pytest.mark.parametrize('batt_const', [-2000, 2000])  # load convention is used for battery in SimSES
def test_energy_cost_reduction(batt_const):
    """Performs a unit test by comparing the expected result for a generic
    time series with the actual result."""

    # set up test data
    n = int(years * 365 * 24 * 60 * 60 / timestep)
    ws_to_kwh = 1/(1000 * 3600)
    unix_timestamp_start_2020 = 1577836800
    time_test = pandas.array([unix_timestamp_start_2020 + timestep * i for i in range(n)])
    load_power = pandas.array([load_const] * n)
    pv_power = pandas.array([pv_const] * n)
    battery_power = pandas.array([batt_const] * n)

    # calculate expected result
    if load_const - pv_const > 0:
        cashflow_base_const = (-load_const + pv_const) * ws_to_kwh * timestep * elec_price
    else:
        cashflow_base_const = (-load_const + pv_const) * ws_to_kwh * timestep * pv_feed
    if load_const - pv_const + batt_const > 0:
        cashflow_battery_const = (-load_const + pv_const - batt_const) * ws_to_kwh * timestep * elec_price
    else:
        cashflow_battery_const = (-load_const + pv_const - batt_const) * ws_to_kwh * timestep * pv_feed
    cashflow_const = cashflow_battery_const - cashflow_base_const
    expected_result = np.array([cashflow_const * n/years] * years)
    # expected_result[-1] = expected_result[-1] - cashflow_const

    # build configs for testing
    gen_sim_config = create_general_config()
    economic_config = create_economic_analysis_config()

    # create data for testing
    energy_management_dict = {EnergyManagementState.TIME: time_test, EnergyManagementState.LOAD_POWER: load_power,
                              EnergyManagementState.PV_POWER: pv_power}
    energy_management_data = EnergyManagementData(gen_sim_config, pandas.DataFrame(energy_management_dict))
    system_dict = {SystemState.TIME: time_test, SystemState.AC_POWER_DELIVERED: battery_power}
    system_data = SystemData(gen_sim_config, pandas.DataFrame(system_dict))

    # get results
    self_consumption_increase_rev_stream = EnergyCostReduction(energy_management_data, system_data, economic_config)
    result1 = self_consumption_increase_rev_stream.get_cashflow()
    assert list(result1.round(0)) == list(expected_result.round(0))


import pandas as pd
import scipy.interpolate

from simses.commons.config.data.battery import BatteryDataConfig
from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.log import Logger
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.cell.electric.properties import ElectricalCellProperties
from simses.technology.lithium_ion.cell.format.abstract import CellFormat
from simses.technology.lithium_ion.cell.format.prismatic import PrismaticCell
from simses.technology.lithium_ion.cell.thermal.properties import ThermalCellProperties
from simses.technology.lithium_ion.cell.type import CellType


class DaimlerLMO(CellType):
    """A LMO (Daimler LMO) is a special cell type and inherited by CellType"""
    """ Source: 
        field data from fcr storage system
    """
    __SOC_HEADER = 'SOC'
    __SOC_IDX = 0
    __DOC_IDX = 0
    __OCV_IDX = 1
    __TEMP_IDX = 1
    __C_RATE_IDX = 0
    __ETA_IDX = 1
    __LENGTH_TEMP_ARRAY = 40
    __LENGTH_SOC_ARRAY = 1001
    __LENGTH_DOC_ARRAY = 1001

    __CELL_VOLTAGE = 3.7  # V
    __CELL_CAPACITY = 50.0  # Ah
    __MAX_VOLTAGE: float = 4.095  # V
    __MIN_VOLTAGE: float = 2.9  # V
    __MIN_TEMPERATURE: float = 248.15  # K
    __MAX_TEMPERATURE: float = 323.15  # K
    __MAX_CHARGE_RATE: float = 2.0  # 1/h
    __MAX_DISCHARGE_RATE: float = 2.0  # 1/h
    __SELF_DISCHARGE_RATE: float = 0.0  # X.X%-soc per day, e.g., 0.015 for 1.5% SOC loss per day
    __MASS: float = 1.7  # kg per cell
    __SPECIFIC_HEAT: float = 823  # J/kgK
    __CONVECTION_COEFFICIENT: float = 15  # W/m2K

    __HEIGHT: float = 113.5  # mm
    __WIDTH: float = 43.8  # mm
    __LENGTH: float = 171.0  # mm

    __COULOMB_EFFICIENCY: float = 1.0  # p.u

    __ELECTRICAL_PROPS: ElectricalCellProperties = ElectricalCellProperties(__CELL_VOLTAGE, __CELL_CAPACITY,
                                                                            __MIN_VOLTAGE, __MAX_VOLTAGE,
                                                                            __MAX_CHARGE_RATE, __MAX_DISCHARGE_RATE,
                                                                            __SELF_DISCHARGE_RATE, __COULOMB_EFFICIENCY)
    __THERMAL_PROPS: ThermalCellProperties = ThermalCellProperties(__MIN_TEMPERATURE, __MAX_TEMPERATURE, __MASS,
                                                                   __SPECIFIC_HEAT, __CONVECTION_COEFFICIENT)
    __CELL_FORMAT: CellFormat = PrismaticCell(__HEIGHT, __WIDTH, __LENGTH)

    # Values from nmc_sanyo_ur18650e (source: https://akkuplus.de/Panasonic-UR18650E-37-Volt-2150mAh-Li-Ion-EOL)

    __USE_FIELD_DATA = False  # use measured data from plant for config (false = lab measurements used)
    __USE_MEASUREMENT = False  # use lab measurement (false = datasheet values)

    if __USE_FIELD_DATA:
        __CELL_RESISTANCE: float = 1.0 * 10 ** (-3)  # Ohm, value from field measurement
    else:
        __CELL_RESISTANCE: float = 0.81 * 10 ** (-3)  # Ohm, value from master's thesis: Gnacy

    def __init__(self, voltage: float, capacity: float, soh: float, battery_config: BatteryConfig,
                 battery_data_config: BatteryDataConfig):
        super().__init__(voltage, capacity, soh, self.__ELECTRICAL_PROPS, self.__THERMAL_PROPS, self.__CELL_FORMAT,
                         battery_config)
        self.__log: Logger = Logger(type(self).__name__)
        rint_file: str = battery_data_config.nmc_molicel_rint_file
        internal_resistance = pd.read_csv(rint_file)  # Ohm
        soc_arr = internal_resistance.iloc[:, self.__SOC_IDX]
        temp_arr = internal_resistance.iloc[:4, self.__TEMP_IDX]
        rint_mat_ch = internal_resistance.iloc[:, 2]
        rint_mat_dch = internal_resistance.iloc[:, 5]
        self.__rint_ch_interp1d = scipy.interpolate.interp1d(soc_arr, rint_mat_ch, kind='linear')
        self.__rint_dch_interp1d = scipy.interpolate.interp1d(soc_arr, rint_mat_dch, kind='linear')
        # TODO these stress factors need to be transferred to the corresponding degradation model
        # capacity_cal_file: str = battery_data_config.nmc_molicel_capacity_cal_file
        # ri_cal_file: str = battery_data_config.nmc_molicel_ri_cal_file
        # capacity_cyc_file: str = battery_data_config.nmc_molicel_capacity_cyc_file
        # ri_cyc_file: str = battery_data_config.nmc_molicel_ri_cyc_file
        # capacity_cyc = pd.read_csv(capacity_cyc_file)  # -
        # capacity_cyc_mat = capacity_cyc.iloc[:self.__LENGTH_DOC_ARRAY, 1]
        # doc_arr = capacity_cyc.iloc[:, self.__DOC_IDX]
        # self.__capacity_cyc_interp1d = scipy.interpolate.interp1d(doc_arr, capacity_cyc_mat, kind='linear')
        # ri_cyc = pd.read_csv(ri_cyc_file)  # -
        # ri_cyc_mat = ri_cyc.iloc[:(self.__LENGTH_DOC_ARRAY + 1), 1]
        # doc_arr = ri_cyc.iloc[:, self.__DOC_IDX]
        # self.__ri_cyc_interp1d = scipy.interpolate.interp1d(doc_arr, ri_cyc_mat, kind='linear')
        # capacity_cal = pd.read_csv(capacity_cal_file)  # -
        # capacity_cal_mat = capacity_cal.iloc[:(self.__LENGTH_TEMP_ARRAY + 1), 2:]
        # soc_arr = capacity_cal.iloc[:, self.__SOC_IDX]
        # temp_arr = capacity_cal.iloc[:(self.__LENGTH_TEMP_ARRAY + 1), self.__TEMP_IDX]
        # self.__capacity_cal_interp1d = scipy.interpolate.interp2d(soc_arr, temp_arr.T, capacity_cal_mat, kind='linear')
        # ri_cal = pd.read_csv(ri_cal_file)  # -
        # ri_cal_mat = ri_cal.iloc[:(self.__LENGTH_TEMP_ARRAY + 1), 2:]
        # soc_arr = ri_cal.iloc[:, self.__SOC_IDX]
        # temp_arr = ri_cal.iloc[:(self.__LENGTH_TEMP_ARRAY + 1), self.__TEMP_IDX]
        # self.__ri_cal_interp1d = scipy.interpolate.interp2d(soc_arr, temp_arr.T, ri_cal_mat, kind='linear')

    def get_open_circuit_voltage(self, battery_state: LithiumIonState) -> float:
        soc: float = battery_state.soc * 100.0
        # Linear OCV fit parameters (field data)
        m = 0.43619
        t = 3.67680

        if self.__USE_MEASUREMENT:  # OCV from measurement MA Gnacy, poly9 fit
            p1 = 7.184106990621546e-16
            p2 = -3.740729945156438e-13
            p3 = 8.302607895987187e-11
            p4 = -1.023080886243304e-08
            p5 = 7.638961638126957e-07
            p6 = -3.543204493824138e-05
            p7 = 0.001009824415000
            p8 = -0.017049050946914
            p9 = 0.165789661369883
            p10 = 2.936578769598463
        else:  # OCV from datasheet, poly9 fit
            p1 = 1.053978763548226e-15
            p2 = -5.472087523147638e-13
            p3 = 1.210062584849170e-10
            p4 = -1.481210133229709e-08
            p5 = 1.091568155357329e-06
            p6 = -4.935612738181030e-05
            p7 = 0.001340177529868
            p8 = -0.020709247233047
            p9 = 0.176208491407495
            p10 = 2.9

        if self.__USE_FIELD_DATA:
            soc = soc / 100
            ocv = m * soc + t
        else:
            ocv = p10 + p9 * soc + p8 * soc ** 2 + p7 * soc ** 3 + p6 * soc ** 4 + p5 * soc ** 5 + \
                    p4 * soc ** 6 + p3 * soc ** 7 + p2 * soc ** 8 + p1 * soc ** 9
        return ocv * self.get_serial_scale()

    def get_internal_resistance(self, battery_state: LithiumIonState) -> float:
        return self.__CELL_RESISTANCE / self.get_parallel_scale() * self.get_serial_scale()

    # TODO these stress factors need to be transferred to the corresponding degradation model
    # def get_stressfkt_ca_cal(self, battery_state: LithiumIonState) -> float:
    #     """
    #     get the stress factor for calendar aging capacity loss
    #
    #     Parameters
    #     ----------
    #     battery_state :
    #
    #     Returns
    #     -------
    #
    #     """
    #     return float(self.__capacity_cal_interp1d(battery_state.soc, battery_state.temperature))
    #
    # def get_stressfkt_ri_cal(self, battery_state: LithiumIonState) -> float:
    #     """
    #     get the stress factor for calendar aging capacity loss
    #
    #     Parameters
    #     ----------
    #     battery_state :
    #
    #     Returns
    #     -------
    #
    #     """
    #     return float(self.__ri_cal_interp1d(battery_state.soc, battery_state.temperature))
    #
    # def get_stressfkt_ca_cyc(self, doc: float) -> float:
    #     return float(self.__capacity_cyc_interp1d(doc))
    #
    # def get_stressfkt_ri_cyc(self, doc: float) -> float:
    #     return float(self.__ri_cyc_interp1d(doc))

    def close(self):
        self.__log.close()



import numpy

from simses.commons.state.system import SystemState
from simses.logic.power_distribution.power_distributor import PowerDistributor
from simses.logic.power_distribution.state import PowerDistributorState


class EfficientPowerDistributor(PowerDistributor):

    """
    EqualPowerDistributor distributes the power equally to all system independent of their current state.
    """

    __SOC_MIN: float = 0.0
    __SOC_MAX: float = 0.99

    __MIN_POWER_SHARE: float = 0.0

    def __init__(self, max_pe_power: float):
        super().__init__()
        self.__pd_states: [PowerDistributorState] = list()
        self.__MIN_POWER: float = max_pe_power * self.__MIN_POWER_SHARE

    def set(self, time: float, states: [SystemState], power) -> None:
        for state in states:
            pd_state: PowerDistributorState = self.__get_pd_state_for(state)
            pd_state.soc = state.soc
            pd_state.max_charge_power = 0.0 if state.soc > self.__SOC_MAX else state.max_charge_power
            pd_state.max_discharge_power = 0.0 if state.soc < self.__SOC_MIN else state.max_discharge_power

    def __get_id_from(self, state: SystemState) -> int:
        return int(state.get(SystemState.SYSTEM_DC_ID))

    def __get_pd_state_for(self, state: SystemState) -> PowerDistributorState:
        id: int = self.__get_id_from(state)
        for pd_state in self.__pd_states:
            if pd_state.sys_id == id:
                return pd_state
        pd_state: PowerDistributorState = PowerDistributorState(sys_id=id)
        self.__pd_states.append(pd_state)
        return pd_state

    def __is_charge(self, power: float) -> bool:
        return power > 0.0

    def get_power_for(self, power: float, state: SystemState) -> float:
        if abs(power) < self.__MIN_POWER:
            return 0.0
        equal_power_distribution: float = abs(power) / len(self.__pd_states)
        is_charge: bool = self.__is_charge(power)
        id: int = self.__get_id_from(state)
        local_power: float = 0.0
        remaining_power_allocation: float = 0.0
        for pd_state in self.__pd_states:
            available_power: float = max(0.0, min(equal_power_distribution, pd_state.max_charge_power if is_charge else pd_state.max_discharge_power))
            remaining_power_allocation += max(0.0, equal_power_distribution - available_power)
            if pd_state.sys_id == id:
                local_power = available_power
        if remaining_power_allocation > 0.0:
            for pd_state in self.__pd_states:
                available_power: float = pd_state.max_charge_power if is_charge else pd_state.max_discharge_power
                power_allocation = max(0.0, min(remaining_power_allocation, available_power - equal_power_distribution))
                remaining_power_allocation -= power_allocation
                if pd_state.sys_id == id:
                    local_power += power_allocation
        return local_power * numpy.sign(power)



from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.cycle_detection.cycle_detector import CycleDetector
from simses.technology.lithium_ion.cell.type import CellType
from simses.technology.lithium_ion.degradation.calendar.nca_generic import NCAGenericCalendarDegradationModel
from simses.technology.lithium_ion.degradation.cyclic.nca_generic import NCAGenericCyclicDegradationModel
from simses.technology.lithium_ion.degradation.degradation_model import DegradationModel


class NCAGenericDegradationModel(DegradationModel):

    def __init__(self, cell_type: CellType, cycle_detector: CycleDetector, battery_config: BatteryConfig):
        super().__init__(cell_type,
                         NCAGenericCyclicDegradationModel(cell_type, cycle_detector),
                         NCAGenericCalendarDegradationModel(cell_type),
                         cycle_detector, battery_config)



from abc import abstractmethod, ABC
from simses.commons.state.abstract_state import State
from simses.system.auxiliary.auxiliary import Auxiliary


class SystemThermalModel(ABC):
    """Thermal model of a storage system"""

    def __init__(self):
        # Air parameters
        self.air_pressure = 1 * 10 ** 5  # Pa, change if container is pressurized
        self.universal_gas_constant = 8.314  # J/kgK
        self.molecular_weight_air = 28.965 * 10 ** -3  # kg/mol

    @abstractmethod
    def get_temperature(self) -> float:
        """
        Returns the current temperature of the system (Internal air temperature)

        Parameters
        -------

        Returns
        -------
        float
            system temperature in Kelvin
        """

        pass

    @abstractmethod
    def get_ol_temperature(self) -> float:
        """
        Returns the current temperature of the outer layer of container

        Parameters
        -------

        Returns
        -------
        float
            temperature in Kelvin
        """

        pass

    @abstractmethod
    def get_il_temperature(self) -> float:
        """
        Returns the current temperature of the inner layer of container

        Parameters
        -------

        Returns
        -------
        float
            temperature in Kelvin
        """

        pass

    @abstractmethod
    def get_ambient_temperature(self) -> float:
        """
        Returns the ambient temperature at the location
        """

        pass

    @abstractmethod
    def get_solar_irradiation_thermal_load(self) -> float:
        """
        returns the solar irradiation thermal load on the container in W
        """
        pass

    @abstractmethod
    def get_hvac_thermal_power(self) -> float:
        """
        returns the total HVAC thermal power in W
        """
        pass

    @abstractmethod
    def get_auxiliaries(self) -> [Auxiliary]:
        pass

    @abstractmethod
    def update_air_parameters(self):
        pass

    @abstractmethod
    def calculate_temperature(self, time: float, storage_system_ac_state: State, storage_system_dc_states: [State]) -> None:
        """
        Calcualtes the current temperature of the system

        Parameters
        ----------
        time : current simulation time
        state : current system state

        Returns
        -------
        :param time:
        :param storage_system_dc_states:
        :param storage_system_ac_state:

        """
        pass

    @abstractmethod
    def reset_profiles(self, ts_adapted: float) -> None:
        """
        Enables looping of the simulation beyond the original length of the time series for the AmbientThermalModel and
        SolarIrradiationModel
        """
        pass

    def close(self):
        """
        Closing all open resources in system thermal model

        Parameters
        ----------

        Returns
        -------

        """
        pass



from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.log import Logger
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.commons.utils.utilities import format_float
from simses.technology.lithium_ion.cell.type import CellType


class BatteryManagementSystem:
    """ BatteryManagementSystem class """

    def __init__(self, cell_type: CellType, battery_config: BatteryConfig):
        self.__log: Logger = Logger(type(self).__name__)
        self.__cell: CellType = cell_type
        self.__min_soc = battery_config.min_soc
        self.__max_soc = battery_config.max_soc

    def update(self, time: float, battery_state: LithiumIonState, power_target: float) -> None:
        """Updating current of lithium_ion state in order to comply with cell type restrictions"""
        battery_state.max_charge_power = self.__get_max_charge_power(time, battery_state)
        battery_state.max_discharge_power = self.__get_max_discharge_power(time, battery_state)
        battery_state.current = self.__get_max_current(battery_state)
        battery_state.voltage = self.__get_voltage(battery_state)
        battery_state.power_loss = self.__get_losses(battery_state)
        self.__set_battery_fulfillment(power_target, battery_state)

    def __get_max_current(self, battery_state: LithiumIonState) -> float:
        if self.__temperature_in_range(battery_state):
            current: float = battery_state.current
            if not self.__current_in_range(battery_state):
                self.__log.warn("Battery current out of range: " + format_float(current))
            voltage: float = battery_state.voltage
            max_charge_current: float = battery_state.max_charge_power / voltage
            max_discharge_current: float = -1.0 * battery_state.max_discharge_power / voltage
            if not max_discharge_current <= current <= max_charge_current:
                msg: str = ''
                msg += 'Current is limited'
                msg += ', current: ' + format_float(current)
                msg += ', max. charge current: ' + format_float(max_charge_current)
                msg += ', max. discharge current: ' + format_float(max_discharge_current)
                self.__log.warn(msg)
            return max(min(current, max_charge_current), max_discharge_current)
        else:
            self.__log.warn("Battery temperature out of range: " + format_float(battery_state.temperature))
            return 0.0

    def __get_voltage(self, battery_state: LithiumIonState) -> float:
        # TODO: combine with __get_max_current
        i = battery_state.current
        r = battery_state.internal_resistance
        ocv = battery_state.voltage_open_circuit
        return ocv + r * i

    def __get_losses(self, battery_state: LithiumIonState) -> float:
        # TODO: combine with __get_max_current
        i = battery_state.current
        r = battery_state.internal_resistance
        return r * i ** 2

    def __get_max_charge_power(self, time: float, battery_state: LithiumIonState):
        voltage: float = battery_state.voltage
        cell_current: float = abs(self.__cell.get_max_current(battery_state))
        soc_current: float = abs(self.__soc_max_charge_current(time, battery_state))
        msg: str = ''
        msg += 'Max charge power'
        msg += ', voltage: ' + format_float(voltage)
        msg += ', cell_current: ' + format_float(cell_current)
        msg += ', soc_current: ' + format_float(soc_current)
        self.__log.debug(msg)
        return min(cell_current, soc_current) * voltage

    def __get_max_discharge_power(self, time: float, battery_state: LithiumIonState):
        voltage: float = battery_state.voltage
        cell_current: float = abs(self.__cell.get_min_current(battery_state))
        soc_current: float = abs(self.__soc_max_discharge_current(time, battery_state))
        msg: str = ''
        msg += 'Max discharge power'
        msg += ', voltage: ' + format_float(voltage)
        msg += ', cell_current: ' + format_float(cell_current)
        msg += ', soc_current: ' + format_float(soc_current)
        self.__log.debug(msg)
        return min(cell_current, soc_current) * voltage

    def __temperature_in_range(self, battery_state: LithiumIonState) -> bool:
        value = battery_state.temperature
        min_value = self.__cell.get_min_temp()
        max_value = self.__cell.get_max_temp()
        return self.__min_max_check(value, min_value, max_value)

    def __voltage_in_range(self, battery_state: LithiumIonState) -> bool:
        # Due to overvoltages this leads to an oscillation of input current and prohibits a full charge of the lithium_ion.
        # TODO implement e.g. linear restricition functions in order to ease current in min/max soc regions
        # return self.__cell.get_min_voltage() <= battery_state.voltage <= self.__cell.get_max_voltage()
        return True

    def __soc_in_range(self, battery_state: LithiumIonState) -> bool:
        value = battery_state.soc  # + RintModel.calculate_dsoc(time, battery_state, self.__cell)
        min_value = self.__min_soc
        max_value = self.__max_soc
        return self.__min_max_check(value, min_value, max_value)

    def __soc_max_charge_current(self, time: float, battery_state: LithiumIonState) -> float:
        denergy = self.__max_soc * battery_state.capacity - battery_state.soe
        current = denergy / battery_state.voltage_open_circuit / (time - battery_state.time) * 3600.0
        msg: str = ''
        msg += 'soc max charge current'
        msg += ', denergy: ' + format_float(denergy)
        msg += ', current: ' + format_float(current)
        self.__log.debug(msg)
        return max(0.0, current)

    def __soc_max_discharge_current(self, time: float, battery_state: LithiumIonState) -> float:
        denergy = battery_state.soe - self.__min_soc * battery_state.capacity
        current = denergy / battery_state.voltage_open_circuit / (time - battery_state.time) * 3600.0
        msg: str = ''
        msg += 'soc max discharge current'
        msg += ', denergy: ' + format_float(denergy)
        msg += ', current: ' + format_float(current)
        self.__log.debug(msg)
        return min(0.0, -current)

    def __current_in_range(self, battery_state: LithiumIonState) -> bool:
        min_value = self.__cell.get_min_current(battery_state)
        max_value = self.__cell.get_max_current(battery_state)
        return self.__min_max_check(battery_state.current, min_value, max_value)

    def __set_battery_fulfillment(self, power_target: float, battery_state: LithiumIonState) -> None:
        power_is = battery_state.current * battery_state.voltage
        if abs(power_is - power_target) < 1e-6:
            battery_state.fulfillment = 1.0
        elif power_target == 0:
            self.__log.error('Power should be 0, but is ' + str(power_is) + ' W - Check BMS function')
            battery_state.fulfillment = 0.0
        else:
            battery_state.fulfillment = abs(power_is / power_target)
            if battery_state.fulfillment < 0 or battery_state.fulfillment > 1:
                self.__log.error('Fulfillment should be between 0 and 1 , but is ' +
                                 str(battery_state.fulfillment) + ' p.u. - Check BMS function')

    @staticmethod
    def __min_max_check(value: float, min_value: float, max_value: float) -> bool:
        return min_value <= value <= max_value

    def close(self) -> None:
        """Closing all resources in lithium_ion management system"""
        self.__log.close()



from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.config.simulation.profile import ProfileConfig
from simses.commons.profile.power.file import FilePowerProfile


class V2GProfile(FilePowerProfile):
    """
    V2GProfile is a specific implementation of FilePowerProfile. It reads a file with a power and time series.
    Values with a positive sign are recognized as a load for the system.

    LoadProfiles require a specific header at the top of the file:\n
    # Annual load consumption in kWh: [Value]
    """

    ANNUAL_CONSUMPTION: str = 'Annual load consumption in kWh'
    """Header of load profile for scaling power values"""

    def __init__(self, profile_config: ProfileConfig, general_config: GeneralSimulationConfig):
        super().__init__(general_config, profile_config.v2g_profile_file,
                         scaling_factor=profile_config.load_scaling_factor)



from simses.commons.state.technology.redox_flow import RedoxFlowState
from simses.technology.redox_flow.degradation.capacity_degradation import \
    CapacityDegradationModel


class NoDegradation(CapacityDegradationModel):
    """Model with no capacity degradation for a redox flow battery."""

    def __init__(self, capacity: float):
        super().__init__(capacity)

    def get_capacity_degradation(self, time: float, redox_flow_state: RedoxFlowState):
        return 0.0

    def close(self):
        super().close()



import pandas

from simses.analysis.data.abstract_data import Data
from simses.analysis.utils import get_mean_for
from simses.analysis.utils import get_positive_values_from, get_sum_for, get_negative_values_from
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.state.technology.redox_flow import RedoxFlowState


class RedoxFlowData(Data):
    """
    Provides time series data from RedoxFlowState
    """

    def __init__(self, config: GeneralSimulationConfig, data: pandas.DataFrame):
        super().__init__(config, data)
        self.__config = config

    @property
    def id(self) -> str:
        return str(int(self._get_first_value(RedoxFlowState.SYSTEM_AC_ID))) + '.' + \
               str(int(self._get_first_value(RedoxFlowState.SYSTEM_DC_ID)))

    @property
    def time(self):
        return self._get_data(RedoxFlowState.TIME)

    @property
    def power(self):
        return self._get_data(RedoxFlowState.POWER)

    @property
    def pump_power(self):
        return self._get_data(RedoxFlowState.PUMP_POWER)

    @property
    def dc_power(self):
        return self.power

    @property
    def energy_difference(self):
        soc = self._get_difference(RedoxFlowState.SOC)
        capacity = self.initial_capacity
        return soc * capacity

    @property
    def soc(self):
        return self._get_data(RedoxFlowState.SOC)

    @property
    def soc_stack(self):
        return self._get_data(RedoxFlowState.SOC_STACK)

    @property
    def capacity(self):
        return self._get_data(RedoxFlowState.CAPACITY) / 1000.0

    @property
    def state_of_health(self):
        return self._get_data(RedoxFlowState.SOH)

    @property
    def resistance(self):
        return self._get_data(RedoxFlowState.INTERNAL_RESISTANCE)

    @property
    def temperature(self):
        return self._get_data(RedoxFlowState.TEMPERATURE)

    @property
    def storage_fulfillment(self):
        return self._get_data(RedoxFlowState.FULFILLMENT)

    @classmethod
    def get_system_data(cls, path: str, config: GeneralSimulationConfig) -> list:
        system_data: [pandas.DataFrame] = cls._get_system_data_for(path, RedoxFlowState, RedoxFlowState.TIME,
                                                                   RedoxFlowState.SYSTEM_AC_ID, RedoxFlowState.SYSTEM_DC_ID)
        res: [RedoxFlowData] = list()
        for data in system_data:
            res.append(RedoxFlowData(config, data))
        return res

    @property
    def current(self):
        return self._get_data(RedoxFlowState.CURRENT)

    @property
    def mean_open_circuit_voltage(self):
        ocv = self._get_data(RedoxFlowState.OPEN_CIRCUIT_VOLTAGE)
        return get_mean_for(ocv)

    @property
    def charge_current_sec(self) -> float:
        """
        Charge energy in kWh (as positive values)

        Returns
        -------

        """
        data = get_positive_values_from(self.current)
        return get_sum_for(data) * self.__config.timestep

    @property
    def discharge_current_sec(self) -> float:
        """
        Discharge energy in kWh (as positive values)

        Returns
        -------

        """
        data = -1 * get_negative_values_from(self.current)
        return get_sum_for(data) * self.__config.timestep

    @property
    def charge_difference(self):
        soc_difference = self._get_difference(RedoxFlowState.SOC)
        charge_total = self.initial_capacity * 1000 * 3600 / self.mean_open_circuit_voltage
        return soc_difference * charge_total





import math

import pandas as pd

from simses.commons.config.data.power_electronics import PowerElectronicsConfig
from simses.commons.log import Logger
from simses.commons.utils.utilities import check
from simses.system.power_electronics.acdc_converter.abstract_acdc_converter import AcDcConverter


class Sinamics120AcDcConverter(AcDcConverter):
    """This class uses a pre-generated Look-up table (LUT) for the efficiency of the Sinamics S120 converter in charging
    and discharging processes. The LUT was generated by Michael Schimpe from measured data using a script written in
    MATLAB. The efficiency is measured for 650 V. Source: https://doi.org/10.1016/j.egypro.2018.11.065 """

    """https://new.siemens.com/global/de/produkte/antriebstechnik/umrichter.html"""

    __VOLUMETRIC_POWER_DENSITY = 143 * 1e6  # W / m3
    __GRAVIMETRIC_POWER_DENSITY = 17000  # W/kg
    __SPECIFIC_SURFACE_AREA = 0.0001  # in m2 / W  # TODO add exact values
    # Exemplary value from:
    # (https://www.iisb.fraunhofer.de/en/research_areas/vehicle_electronics/dcdc_converters/High_Power_Density.html)
    # ( https://www.apcuk.co.uk/app/uploads/2018/02/PE_Full_Pack.pdf )

    def __init__(self, max_power: float, config: PowerElectronicsConfig):
        super().__init__(max_power)
        __filename = config.sinamics_efficiency_file
        self.__efficiency = pd.read_csv(__filename)  # pu
        self.__log: Logger = Logger(type(self).__name__)

    def to_ac(self, power: float, voltage: float) -> float:
        check(power)
        res: float = 0.0
        if power < 0.0:
            power_factor = abs(power) / self.max_power
            res = power / self.__get_efficiency(power, power_factor)
        return res

    def to_dc(self, power: float, voltage: float) -> float:
        check(power)
        res: float = 0.0
        if power > 0.0:
            power_factor = abs(power) / self.max_power
            res = power * self.__get_efficiency(power, power_factor)
        return res

    def to_dc_reverse(self, dc_power: float, voltage: float) -> float:
        check(dc_power)
        res: float = 0.0
        if dc_power > 0.0:
            power_factor = min(abs(dc_power) / (self.max_power * self.__efficiency.iloc[-1, 0]), 1.0)
            res = dc_power / self.__get_efficiency(dc_power, power_factor)
        return res

    def to_ac_reverse(self, dc_power: float, voltage: float) -> float:
        check(dc_power)
        res: float = 0.0
        if dc_power < 0.0:
            power_factor = min(abs(dc_power) / (self.max_power * self.__efficiency.iloc[-1, 1]), 1.0)
            res = dc_power * self.__get_efficiency(dc_power, power_factor)
        return res

    def __get_efficiency(self, power: float, power_factor: float) -> float:
        if power_factor < 0.0 or power_factor > 1.0:
            raise Exception('Power factor is not possible: ' + str(power_factor))
        power_factor = int(round(power_factor * 1000))
        if power_factor > 0:
            if power > 0: # charging
                return self.__efficiency.iloc[power_factor, 0]
            elif power < 0: # discharging
                return self.__efficiency.iloc[power_factor, 1]
        else:
            return math.inf

    @property
    def volume(self) -> float:
        return self.max_power / self.__VOLUMETRIC_POWER_DENSITY

    @property
    def mass(self):
        return self.max_power / self.__GRAVIMETRIC_POWER_DENSITY

    @property
    def surface_area(self) -> float:
        return self.max_power * self.__SPECIFIC_SURFACE_AREA

    @classmethod
    def create_instance(cls, max_power: float, power_electronics_config: PowerElectronicsConfig):
        return Sinamics120AcDcConverter(max_power, power_electronics_config)

    def close(self) -> None:
        pass



from configparser import ConfigParser

from simses.commons.data.data_handler import DataHandler
from simses.commons.state.technology.electrolyzer import ElectrolyzerState
from simses.system.auxiliary.gas_drying.gas_dryer import GasDrying
from simses.system.auxiliary.gas_drying.hydrogen import HydrogenGasDrying
from simses.system.auxiliary.pump.abstract_pump import Pump
from simses.system.auxiliary.pump.fixeta_centrifugal import FixEtaCentrifugalPump
from simses.system.auxiliary.water_heating.water_heater import WaterHeating
from simses.technology.hydrogen.electrolyzer.degradation.degradation_model import DegradationModel
from simses.technology.hydrogen.electrolyzer.factory import ElectrolyzerFactory
from simses.technology.hydrogen.electrolyzer.pressure.abstract_model import PressureModel
from simses.technology.hydrogen.electrolyzer.stack.no_electrolyzer import NoElectrolyzer
from simses.technology.hydrogen.electrolyzer.stack.stack_model import ElectrolyzerStackModel
from simses.technology.hydrogen.electrolyzer.thermal.abstract_model import ThermalModel


class Electrolyzer:
    """
    Electrolyzer is the top level class incorporating a ElectrolyzerStackModel, a ThermalModel and a PressureModel. The
    specific classes are instantiated within the ElectrolyzerFactory.
    """

    __PUMP_EFFICIENCY: float = 0.7

    def __init__(self, system_id: int, storage_id: int, electrolyzer: str, max_power: float,
                 temperature: float, config: ConfigParser, data_handler: DataHandler):
        super().__init__()
        self.__data_handler: DataHandler = data_handler
        self.__pump: Pump = FixEtaCentrifugalPump(self.__PUMP_EFFICIENCY)
        self.__gas_drying: GasDrying = HydrogenGasDrying()
        self.__water_heating: WaterHeating = WaterHeating()
        factory: ElectrolyzerFactory = ElectrolyzerFactory(config)
        self.__stack_model: ElectrolyzerStackModel = factory.create_electrolyzer_stack(electrolyzer, max_power)
        self.__pressure_model: PressureModel = factory.create_pressure_model(self.__stack_model)
        self.__thermal_model: ThermalModel = factory.create_thermal_model(self.__stack_model, self.__water_heating,
                                                                          self.__pump, temperature)
        self.__degradation_model: DegradationModel = factory.create_degradation_model(self.__stack_model)
        self.__state: ElectrolyzerState = factory.create_state(system_id, storage_id, temperature,
                                                               self.__stack_model, self.__pressure_model)
        factory.close()
        self.__log_data: bool = not isinstance(self.__stack_model, NoElectrolyzer)
        self.__write_data()

    def update(self, time: float, power: float) -> None:
        """
        Updates hydrogen states that are corrolated with the electrolyzer such as current, voltage, hydrogen production,
        water use and temperature

        Parameters
        ----------
        time :
        hydrogen_state :
        power :

        Returns
        -------

        """
        local_power = power if power > 0.0 else 0.0
        state = self.__state
        self.__stack_model.update(local_power, state)
        state.power = local_power  # TODO: is at this point clear if the electrolyzer can take all of the supplied power?

        # state values that will be needed before they were updated
        pressure_cathode_0 = state.pressure_cathode  # barg
        pressure_anode_0 = state.pressure_anode  # barg

        # update temperature and pressure of the stack
        self.__pressure_model.update(time, state)
        state.water_use += state.water_outflow_cathode + state.water_outflow_anode
        self.__thermal_model.update(time, state, pressure_cathode_0, pressure_anode_0)
        # self.convection_heat_el = self.__thermal_model.get_convection_heat()

        # update degradation of stack
        state.reference_voltage = self.__stack_model.get_reference_voltage_eol(state.resistance_increase,
                                                                                  state.exchange_current_decrease)
        self.__degradation_model.update(time, state)

        # update auxilliary losses
        state.power_water_heating = self.__thermal_model.get_power_water_heating()
        # pressure_loss: float = self.__thermal_model.get_tube_pressure_loss() * state.water_flow * \
        #                        ConstantsHydrogen.MOLAR_MASS_WATER / ConstantsHydrogen.DENSITY_WATER
        # self.__pump.calculate_pump_power(pressure_loss)
        state.power_pump = self.__pump.get_pump_power()
        self.__gas_drying.calculate_gas_drying_power(pressure_cathode_0+1, state.hydrogen_outflow)
        state.power_gas_drying = self.__gas_drying.get_gas_drying_power()
        state.time = time
        self.__write_data()

    def __write_data(self) -> None:
        if self.__log_data:
            self.__data_handler.transfer_data(self.__state.to_export())

    def get_auxiliaries(self):
        return [self.__gas_drying, self.__water_heating, self.__pump]

    @property
    def state(self) -> ElectrolyzerState:
        return self.__state

    def close(self):
        self.__stack_model.close()
        self.__pressure_model.close()
        self.__thermal_model.close()
        self.__degradation_model.close()
        self.__pump.close()
        self.__water_heating.close()



# Following requirements are needed:
# 1) Install pylint from pip
# 2) Install graphviz (https://graphviz.org/)
#       Important: Add graphviz to PATH variable!
#       First time you run graphviz you need to run 'dot -c' as admin!
#       Maybe you have to restart your PC in between.
# Within pylint the tool pyreverse exist. For any help call 'pyreverse -h'.
# With pyreverse you are able to generate graphs from your code.
import os
from os.path import basename, dirname
from subprocess import check_call

from simses import main
from simses.analysis.evaluation.economic.revenue_stream import abstract_revenue_stream
from simses.analysis.evaluation.technical import technical_evaluation
from simses.commons import config, state
from simses.commons.state import parameters
from simses.logic.energy_management.strategy import operation_strategy, operation_priority
from simses.system.auxiliary import auxiliary
from simses.system.auxiliary.heating_ventilation_air_conditioning import fix_cop_hvac_pid_control
from simses.system.housing import abstract_housing
from simses.system.power_electronics.acdc_converter import abstract_acdc_converter
from simses.system.power_electronics.dcdc_converter import abstract_dcdc_converter
from simses.system.test import test_fix_cop_hvac, test_acdc_converter, test_location_ambient_temperature, \
    test_constant_ambient_temperature
from simses.system.thermal.ambient import ambient_thermal_model
from simses.system.thermal.model import system_thermal_model
from simses.technology.lithium_ion.cell import type
from simses.technology.lithium_ion.degradation import degradation_model
from simses.technology.lithium_ion.test import test_cell_type


class Filter:
    PUB_ONLY: str = 'PUB_ONLY'
    ALL: str = 'ALL'
    SPECIAL: str = 'SPECIAL'
    OTHER: str = 'OTHER'


class Output:
    PDF: str = 'pdf'
    PNG: str = 'png'
    DOT: str = 'dot'


def is_output(file: str) -> bool:
    return file.endswith(Output.PDF) or \
           file.endswith(Output.PNG) or \
           file.endswith(Output.DOT)


def create_ignore_list() -> str:
    ignore: [str] = list()
    ignore.append(test_acdc_converter)
    ignore.append(operation_priority)
    ignore.append(test_cell_type)
    ignore.append(parameters)
    ignore.append(test_constant_ambient_temperature)
    ignore.append(test_location_ambient_temperature)
    ignore.append(test_fix_cop_hvac)
    ignore.append(fix_cop_hvac_pid_control)
    res: str = ''
    for module in ignore:
        res += basename(module.__file__) + ','
    return res


def create_call_for(package: str, ignore: str, cls: str = None) -> [str]:
    call: [str] = list()
    call.append('pyreverse')
    call.append('-o' + Output.PDF)
    call.append('-f' + Filter.PUB_ONLY)
    call.append('-mn')  # do not use module name for classes, my for yes
    call.append('-k')  # show only classnames
    call.append('--ignore=' + ignore)  # files not to be parsed
    if cls is not None:
        call.append('-c' + cls)  # create a class diagram
        # call.append('-s 1')  # depth of associated classes
        # call.append('-a 1')  # depth of ancestors
    call.append('-p' + basename(package))  # name output of basename from package
    call.append(package)
    return call


def create_graph_for(file: str, ignore: str, cls: str = None) -> None:
    print('Creating graph for ' + file)
    call: [str] = create_call_for(file, ignore, cls)
    check_call(call, cwd=dirname(__file__))


def remove_all_graphs() -> None:
    print('Removing old graphs...')
    directory: str = dirname(__file__)
    for item in os.listdir(directory):
        file = os.path.join(directory, item)
        if os.path.isfile(file) and is_output(file):
            os.remove(file)


packages: [str] = list()
packages.append(abstract_dcdc_converter)
packages.append(abstract_acdc_converter)
packages.append(ambient_thermal_model)
packages.append(operation_strategy)
packages.append(type)
packages.append(config)
packages.append(state)
packages.append(abstract_revenue_stream)
packages.append(technical_evaluation)
packages.append(auxiliary)
packages.append(system_thermal_model)
packages.append(degradation_model)
packages.append(abstract_housing)
packages.append(main)

modules: dict = dict()
# modules.update({degradation_model: DegradationModel})

remove_all_graphs()

ignore: str = create_ignore_list()

for package in packages:
    create_graph_for(dirname(package.__file__), ignore)

for module, cls in modules.items():
    create_graph_for(module.__file__, ignore, cls.__name__)

# Conversion from dot files to pdf
# from graphviz import render
# render(Output.DOT, Output.PDF, 'classes.dot')
# render(Output.DOT, Output.PDF, 'packages.dot')



from simses.commons.config.data.electrolyzer import ElectrolyzerDataConfig
from simses.commons.log import Logger
from simses.commons.state.technology.electrolyzer import ElectrolyzerState
from simses.technology.hydrogen.electrolyzer.stack.pem_analytic.efficiency_curves import \
    PemElectrolyzerEfficiencyCurves
from simses.technology.hydrogen.electrolyzer.stack.pem_analytic.electrical_model import \
    PemElectricalModel
from simses.technology.hydrogen.electrolyzer.stack.pem_analytic.fluid_model import \
    PemFluidModel
from simses.technology.hydrogen.electrolyzer.stack.pem_analytic.membrane_model import \
    PemMembraneModel
from simses.technology.hydrogen.electrolyzer.stack.pem_analytic.parameters import \
    ParametersPemElectrolyzerMultiDimAnalytic
from simses.technology.hydrogen.electrolyzer.stack.pem_analytic.pressure_model import \
    PemPressureModel
from simses.technology.hydrogen.electrolyzer.stack.stack_model import ElectrolyzerStackModel


class PemElectrolyzerMultiDimAnalytic(ElectrolyzerStackModel):
    """An PEM-Electrolyzer is a special typ of an Electrolyzer"""

    # TODO Are those variables the same for every PemElectrolyzerMultiDimAnalytic?

    __NOM_POWER_CELL = 5000  # W   side length = 35 cm
    __MAX_POWER_CELL = 7962.5  # W  based on polarizationscurve model from: PEM-Elektrolyse-Systeme zur Anwendung in Power-to-Gas Anlagen
    __RATIO_NOM_TO_MAX = __NOM_POWER_CELL / __MAX_POWER_CELL

    __NOMINAL_CURRENT_DENSITY = 2  # A/cm²
    __MAX_CURRENTDENSITY_CELL = 3  # A/cm2
    __NOM_CURRENTDENSITY_CELL = 2  # A/cm²
    __MIN_CURRENTDENSITY_CELL = 0  # A/cm2
    __NOM_VOLTAGE_CELL = 2  # V

    __HEAT_CAPACITY = 0.448  #  J/K/W for >= 50 kW from: Maximilian Möckl ZAE
    __WATER_POWER_DENSITY = 0.025  # kg / kW from: Maximilian Möckl ZAE

    def __init__(self, electrolyzer_maximal_power: float, electrolyzer_data_config: ElectrolyzerDataConfig):
        super().__init__()
        self.__log: Logger = Logger(type(self).__name__)
        self.__number_cells: float = max(round(electrolyzer_maximal_power / self.__MAX_POWER_CELL), 1)
        self.__max_stack_power: float = self.__number_cells * self.__MAX_POWER_CELL  # W
        self.__nominal_stack_power: float = self.__RATIO_NOM_TO_MAX * self.__max_stack_power  # W
        self.__heat_capacity_stack: float = self.__HEAT_CAPACITY * self.__nominal_stack_power  # J/K
        self.__water_in_stack: float = self.__WATER_POWER_DENSITY * self.__nominal_stack_power / 1000.0  # 0.125 kg/kW
        self.__log.info('maximal stack power of electrolyzer was adapted to: ' + str(self.__max_stack_power))
        self.__log.info('nominal stack power of electrolyzer is: ' + str(self.__nominal_stack_power))
        self.__log.info('number of serial cells is: ' + str(self.__number_cells))
        # initialize models
        self.__parameters = ParametersPemElectrolyzerMultiDimAnalytic(electrolyzer_data_config)
        self.__pressure_model: PemPressureModel = PemPressureModel(self.__parameters)
        self.__membrane_model: PemMembraneModel = PemMembraneModel(self.__pressure_model)
        self.__electrical_model: PemElectricalModel = PemElectricalModel(self.__membrane_model, self.__pressure_model,
                                                                         self.__parameters)
        self.__fluid_model: PemFluidModel = PemFluidModel(self.__membrane_model)
        self.__efficiency_curve_calculation = PemElectrolyzerEfficiencyCurves(self.__electrical_model,
                                                                              self.__membrane_model,
                                                                              self.__pressure_model, self.__fluid_model)
        # initialize variables
        self.__voltage_stack: float = 0.0
        self.__current_stack: float = 0.0
        self.__current_cell: float = 0.0
        self.__h2o_use_stack: float = 0.0
        self.__o2_generation_stack: float = 0.0
        self.__h2_generation_stack: float = 0.0
        self.__part_pressure_h2: float = 0.0
        self.__part_pressure_o2: float = 0.0
        self.__sat_pressure_h2o: float = 0.0

    def calculate(self, power: float, state: ElectrolyzerState):
        # calculation of current, partial pressures and cell voltage at cell level
        power_density: float = power / self.__number_cells / self.__membrane_model.get_geometric_area_cell()  # W/cm2
        # calculation of current and voltage at stack level
        current_dens_cell = self.__electrical_model.get_current_density(power_density, state)  # A/cm2
        voltage_cell = self.__electrical_model.get_cell_voltage(current_dens_cell, state)  # V
        self.__current_stack = current_dens_cell * self.__membrane_model.get_geometric_area_cell()  # A  cells are serial connected to a Stack
        self.__current_cell = self.__current_stack
        self.__voltage_stack = voltage_cell * self.__number_cells  # V  cells are serial connected to one stack
        # calculation of pressures
        self.__part_pressure_h2 = self.__pressure_model.get_partial_pressure_h2(state, current_dens_cell)
        self.__part_pressure_o2 = self.__pressure_model.get_partial_pressure_o2(state, current_dens_cell)
        self.__sat_pressure_h2o = self.__pressure_model.get_sat_pressure_h2o(state.temperature)
        # net use of water, production of hydrogen and oxygen at Stack level
        h2_net_cathode = self.__fluid_model.get_h2_net_cathode(state, self.__current_cell)  # mol/s
        o2_net_anode = self.__fluid_model.get_o2_net_anode(state, self.__current_cell)  # mol/s
        h2o_net_use_cell: float = self.__fluid_model.get_h2o_net_use_cell(state, self.__current_cell)
        self.__h2o_use_stack = h2o_net_use_cell * self.__number_cells  # mol/s
        self.__o2_generation_stack = o2_net_anode * self.__number_cells  # mol/s
        self.__h2_generation_stack = h2_net_cathode * self.__number_cells  # mol/s

    def get_nominal_current_density(self):
        return self.__NOMINAL_CURRENT_DENSITY

    def get_reference_voltage_eol(self, resistance_increase, exchange_current_decrease) -> float:
        """
        return cell voltage of electrolyzer at 2 A/cm², p_anode = 0 barg, p_cathode = 0 barg and temperature = 80°C
        this cell voltage is needed for the calculation of the SOH of the electrolyzer
        :return: cell voltage in V
        """
        # TODO implement function
        # return self.calculate_cell_voltage(30, 0, 0, 2, resistance_increase, exchange_current_decrease)
        ref_electrolyzer_state = ElectrolyzerState(1, 1)
        ref_electrolyzer_state.temperature = 80  # °C
        ref_electrolyzer_state.pressure_cathode = 30  # barg
        ref_electrolyzer_state.pressure_anode = 0  # barg
        ref_electrolyzer_state.exchange_current_decrease = exchange_current_decrease
        ref_electrolyzer_state.resistance_increase = resistance_increase
        current_dens_ref = 2  # A/cm²
        return self.__electrical_model.get_cell_voltage(current_dens_ref, ref_electrolyzer_state)

    def get_current(self):
        return self.__current_stack

    def get_current_density(self):
        return self.__current_stack / self.__membrane_model.get_geometric_area_cell()

    def get_hydrogen_production(self):
        return self.__h2_generation_stack

    def get_oxygen_production(self):
        return self.__o2_generation_stack

    def get_voltage(self):
        return self.__voltage_stack

    def get_water_use(self):
        return self.__h2o_use_stack

    def get_number_cells(self):
        return self.__number_cells

    def get_geom_area_stack(self):
        return self.__number_cells * self.__membrane_model.get_geometric_area_cell()

    def get_nominal_stack_power(self):
        return self.__nominal_stack_power

    def get_heat_capacity_stack(self):
        return self.__heat_capacity_stack

    def get_water_in_stack(self):
        return self.__water_in_stack

    def get_partial_pressure_h2(self):
        return self.__part_pressure_h2

    def get_partial_pressure_o2(self):
        return self.__part_pressure_o2

    def get_sat_pressure_h2o(self):
        return self.__sat_pressure_h2o

    def get_efficiency_curve(self, hydrogen_state: ElectrolyzerState):
        return self.__efficiency_curve_calculation.calculate_efficiency_curves(hydrogen_state)

    def close(self):
        self.__log.close()

    def get_thermal_resistance_stack(self):
        pass



from abc import ABC, abstractmethod

from simses.commons.state.technology.electrolyzer import ElectrolyzerState


class ThermalModel(ABC):

    def __init__(self):
        self.__max_water_flow_stack = 0  # mol/s/cell

    def update(self, time: float, state: ElectrolyzerState, pressure_cathode_0, pressure_anode_0) -> None:
        """Updating temperature of electrolyzer (°C) stack in hydrogen state"""
        self.calculate(time, state, pressure_cathode_0, pressure_anode_0)
        state.water_flow = self.get_water_flow_stack()
        self.calculate_pump_power(state.water_flow)
        # self.calculate_tube_pressure_loss(state.water_flow)
        state.temperature = self.get_temperature()

    @abstractmethod
    def calculate(self, time: float, state: ElectrolyzerState, pressure_cathode_0, pressure_anode_0) -> None:
        pass

    @abstractmethod
    def get_temperature(self) -> float:
        pass

    @abstractmethod
    def get_water_flow_stack(self) -> float:
        pass

    @abstractmethod
    def get_power_water_heating(self) -> float:
        pass

    @abstractmethod
    def calculate_pump_power(self, water_flow_stack: float) -> None:
        pass

    # @abstractmethod
    # def calculate_tube_pressure_loss(self, water_flow_stack: float) -> None:
    #     pass

    @abstractmethod
    def get_pump_power(self) -> float:
        pass

    # @abstractmethod
    # def get_tube_pressure_loss(self) -> float:
    #     pass

    @abstractmethod
    def get_convection_heat(self) -> float:
        pass

    def close(self) -> None:
        """Closing all resources in thermal model"""
        pass



import numpy as np

from simses.commons.log import Logger
from simses.commons.utils.utilities import check
from simses.system.power_electronics.acdc_converter.abstract_acdc_converter import AcDcConverter


class NottonAcDcConverter(AcDcConverter):
    """  Notton, G.; Lazarov, V.; Stoyanov, L. (2010): Optimal sizing of
           a grid-connected PV system for various PV module technologies
           and inclinations, inverter efficiency characteristics and
           locations. In: Renewable Energy 35 (2), S. 541-554.
           DOI: 10.1016/j.renene.2009.07.013. """

    # Notton Type 1 inverter
    # __P0 = 0.0145
    # __K = 0.0437

    # Notton Type 2 inverter
    __P0 = 0.0072
    __K = 0.0345

    # Notton Type 3 inverter
    # __P0 = 0.0088
    # __K = 0.1149

    __VOLUMETRIC_POWER_DENSITY = 143 * 1e6  # W / m3
    __GRAVIMETRIC_POWER_DENSITY = 17000  # W/kg
    __SPECIFIC_SURFACE_AREA = 0.0001  # in m2 / W  # TODO add exact values
    # Exemplary value from:
    # (https://www.iisb.fraunhofer.de/en/research_areas/vehicle_electronics/dcdc_converters/High_Power_Density.html)
    # ( https://www.apcuk.co.uk/app/uploads/2018/02/PE_Full_Pack.pdf )

    def __init__(self, max_power: float):
        super().__init__(max_power)
        self.__log: Logger = Logger(type(self).__name__)

    def to_ac(self, power: float, voltage: float) -> float:
        check(power)
        res: float = 0.0
        if power < 0.0:
            res = power / self.__get_efficiency(power)
        return res

    def to_dc(self, power: float, voltage: float) -> float:
        check(power)
        res: float = 0.0
        if power > 0.0:
            res = power * self.__get_efficiency(power)
        return res

    def to_dc_reverse(self, dc_power: float, voltage: float) -> float:
        check(dc_power)
        res: float = 0.0
        if dc_power > 0.0:
            p = - dc_power / (1 - dc_power * self.__K / self.max_power)
            q = - self.__P0 * dc_power / (1 / self.max_power - abs(dc_power) * self.__K / self.max_power ** 2)
            res = max(0.0, -p / 2 + np.sqrt((p / 2) ** 2 - q))
        return res

    def to_ac_reverse(self, dc_power: float, voltage: float) -> float:
        check(dc_power)
        res: float = 0.0
        if dc_power < 0.0:
            p = self.max_power / self.__K
            q = (self.__P0 * self.max_power ** 2 - abs(dc_power) * self.max_power) / self.__K
            res = min(0.0, -(-p / 2 + np.sqrt((p / 2) ** 2 - q)))
        return res

    def __get_efficiency(self, power: float) -> float:
        power_factor = abs(power) / self.max_power
        if power_factor < 0.0 or power_factor > 1.0:
            raise Exception('Power factor is not possible: ' + str(power_factor))
        return power_factor / (power_factor + self.__P0 + self.__K * power_factor ** 2)

    @property
    def volume(self) -> float:
        return self.max_power / self.__VOLUMETRIC_POWER_DENSITY

    @property
    def mass(self):
        return self.max_power / self.__GRAVIMETRIC_POWER_DENSITY

    @property
    def surface_area(self) -> float:
        return self.max_power * self.__SPECIFIC_SURFACE_AREA

    @classmethod
    def create_instance(cls, max_power: float, power_electronics_config=None):
        return NottonAcDcConverter(max_power)

    def close(self) -> None:
        self.__log.close()



from simses.commons.state.technology.electrolyzer import ElectrolyzerState


class PemPressureModel:

    A_CATHODE = 2.4  # bar cm²/A
    A_ANODE = 2.8  # bar cm²/A

    def __init__(self, parameters):
        self.__parameters = parameters
        #self.__sat_pressure_h2o: float = 0.03189409713071401  # bar
        #self.__part_pressure_h2: float = 1.0 - self.__sat_pressure_h2o  # bar
        #self.__part_pressure_o2: float = 1.0 - self.__sat_pressure_h2o  # bar

    def get_sat_pressure_h2o(self, stack_temperature: float) -> float:
            return self.__get_sat_pressure_h2o(stack_temperature)

    def get_sat_pressure_h2o_for_current_calc(self, stack_temperature: float):
        return self.__get_sat_pressure_h2o(stack_temperature, self.__parameters.q10, self.__parameters.q11,
                                           self.__parameters.q12, self.__parameters.q13, self.__parameters.q14)

    def __get_sat_pressure_h2o(self, stack_temperature: float, q10: float=1, q11: float=1, q12: float=1, q13: float=1,
                               q14: float=1) -> float:
        """ stack temperature needs to be in °C """
        return q10 * 10 ** (- q11 * 2.1794 + q12 * 0.02953 *
                           stack_temperature - q13 * 9.1837 * 10 ** (-5) *
                           stack_temperature ** 2 + q14 * 1.4454 * 10 ** (-7) *
                           stack_temperature ** 3)  # bar saturation pressure of water

    def get_partial_pressure_h2_for_current_calc(self, state: ElectrolyzerState, current_dens_cell: float):
        return self.__get_partial_pressure_h2(state.pressure_cathode, current_dens_cell,
                                              self.get_sat_pressure_h2o_for_current_calc(state.temperature),
                                              self.__parameters.q15)

    def get_partial_pressure_h2(self, state: ElectrolyzerState, current_dens_cell: float):
        return self.__get_partial_pressure_h2(state.pressure_cathode, current_dens_cell,
                                              self.get_sat_pressure_h2o(state.temperature))

    def __get_partial_pressure_h2(self, p_cathode: float, current_dens_cell: float, sat_pressure_h2o: float,
                                  q11: float=1) -> float:
        return (1 + q11 * p_cathode) - sat_pressure_h2o + self.A_CATHODE * current_dens_cell # bar

    def get_partial_pressure_o2_for_current_calc(self, state: ElectrolyzerState, current_dens_cell: float):
        return self.__get_partial_pressure_o2(state.pressure_anode, current_dens_cell,
                                              self.get_sat_pressure_h2o_for_current_calc(state.temperature),
                                              self.__parameters.q9)

    def get_partial_pressure_o2(self, state: ElectrolyzerState, current_dens_cell: float):
         return self.__get_partial_pressure_o2(state.pressure_anode, current_dens_cell,
                                               self.get_sat_pressure_h2o(state.temperature))

    def __get_partial_pressure_o2(self, p_anode: float, current_dens_cell: float, sat_pressure_h2o: float,
                                  q12: float=1) -> float:
        return (1 + q12 * p_anode) - sat_pressure_h2o + self.A_ANODE * current_dens_cell  # bar



from simses.commons.config.data.battery import BatteryDataConfig
from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.cycle_detection.cycle_detector import CycleDetector
from simses.technology.lithium_ion.cell.type import CellType
from simses.technology.lithium_ion.degradation.calendar.nmc_molicel import \
    MolicelNMCCalendarDegradationModel
from simses.technology.lithium_ion.degradation.cyclic.nmc_molicel import \
    MolicelNMCCyclicDegradationModel
from simses.technology.lithium_ion.degradation.degradation_model import DegradationModel


class DaimlerLMODegradationModel(DegradationModel):
    """ Degradation Model for Second-Life cell. Specific model unknown, therefore using different existing
    degradation models from SimSES."""

    def __init__(self, cell_type: CellType, cycle_detector:CycleDetector, battery_config: BatteryConfig, battery_data_config: BatteryDataConfig):
        super().__init__(cell_type, MolicelNMCCyclicDegradationModel(cell_type, cycle_detector, battery_data_config),
                         MolicelNMCCalendarDegradationModel(cell_type, battery_data_config), cycle_detector,
                         battery_config, initial_degradation_possible=True)



from abc import abstractmethod

from simses.system.auxiliary.auxiliary import Auxiliary


class GasDrying(Auxiliary):
    """ GasDrying is an auxiliary. It calculates the energy that is needed for drying the product gas according to the
    specified drying level"""

    def __init__(self):
        super().__init__()

    def ac_operation_losses(self) -> float:
        """
        Calculates the ac operation losses
        :return:
         float:
            Power in W
        """
        return self.get_gas_drying_power()

    @abstractmethod
    def calculate_gas_drying_power(self, pressure_cathode, h2_outflow) -> None:
        pass

    @abstractmethod
    def get_gas_drying_power(self) -> float:
        pass


import math
import numpy

from simses.analysis.data.redox_flow import RedoxFlowData
from simses.analysis.evaluation.plotting.axis import Axis
from simses.analysis.evaluation.plotting.plotly_plotting import PlotlyPlotting
from simses.analysis.evaluation.plotting.plotter import Plotting
from simses.analysis.evaluation.result import EvaluationResult, Description, Unit
from simses.analysis.evaluation.technical.technical_evaluation import TechnicalEvaluation
from simses.commons.config.analysis.general import GeneralAnalysisConfig
from simses.commons.log import Logger
from simses.commons.state.technology.redox_flow import RedoxFlowState
from simses.commons.utils.utilities import format_float


class RedoxFlowTechnicalEvaluation(TechnicalEvaluation):

    title_overview = 'Redox flow results'
    title_soc = 'Redox flow SOC difference'
    title_temperature = 'Redox flow temperature'

    def __init__(self, data: RedoxFlowData, config: GeneralAnalysisConfig, path: str):
        super().__init__(data, config)
        self.__log: Logger = Logger(type(self).__name__)
        title_extension: str = ' for system ' + self.get_data().id
        self.title_overview += title_extension
        self.__result_path = path

    def evaluate(self):
        super().evaluate()
        self.append_result(EvaluationResult(Description.Technical.EQUIVALENT_FULL_CYCLES, Unit.NONE,
                                            self.equivalent_full_cycles))
        self.append_result(EvaluationResult(Description.Technical.DEPTH_OF_DISCHARGE, Unit.PERCENTAGE,
                                            self.depth_of_discharges))
        self.append_result(EvaluationResult(Description.Technical.COULOMB_EFFICIENCY, Unit.PERCENTAGE,
                                            self.coulomb_efficiency))
        self.print_results()

    def plot(self) -> None:
        self.overview_plotting()
        self.soc_comparison_plotting()
        self.temperature_plotting()

    def overview_plotting(self) -> None:
        data: RedoxFlowData = self.get_data()
        plot: Plotting = PlotlyPlotting(title=self.title_overview, path=self.__result_path)
        xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=RedoxFlowState.TIME)
        yaxis: [Axis] = list()
        yaxis.append(Axis(data=data.resistance, label=RedoxFlowState.INTERNAL_RESISTANCE,
                          color=PlotlyPlotting.Color.RESISTANCE_BLACK))
        yaxis.append(Axis(data=data.capacity, label='C in kWh', color=PlotlyPlotting.Color.SOH_GREEN))
        yaxis.append(Axis(data=data.pump_power, label=RedoxFlowState.PUMP_POWER,
                          color=PlotlyPlotting.Color.AC_POWER_BLUE, linestyle=PlotlyPlotting.Linestyle.DASH_DOT))
        yaxis.append(Axis(data=data.power, label=RedoxFlowState.POWER, color=PlotlyPlotting.Color.DC_POWER_GREEN))
        plot.subplots(xaxis=xaxis, yaxis=yaxis)
        self.extend_figures(plot.get_figures())

    def soc_comparison_plotting(self) -> None:
        data: RedoxFlowData = self.get_data()
        plot: Plotting = PlotlyPlotting(title=self.title_soc, path=self.__result_path)
        xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=RedoxFlowState.TIME)
        yaxis: [Axis] = list()
        yaxis.append(Axis(data=data.soc, label=RedoxFlowState.SOC, color=PlotlyPlotting.Color.SOC_BLUE,
                          linestyle=PlotlyPlotting.Linestyle.SOLID))
        yaxis.append(Axis(data=data.soc_stack, label=RedoxFlowState.SOC_STACK, color=PlotlyPlotting.Color.STACK_PURPLE,
                          linestyle=PlotlyPlotting.Linestyle.DOTTED))
        plot.lines(xaxis, yaxis)
        self.extend_figures(plot.get_figures())

    def temperature_plotting(self) -> None:
        data: RedoxFlowData = self.get_data()
        plot: Plotting = PlotlyPlotting(title=self.title_temperature, path=self.__result_path)
        xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=RedoxFlowState.TIME)
        yaxis: [Axis] = list()
        yaxis.append(Axis(data=data.temperature, label=RedoxFlowState.TEMPERATURE,
                          color=PlotlyPlotting.Color.TEMPERATURE_RED, linestyle=PlotlyPlotting.Linestyle.SOLID))
        plot.lines(xaxis, yaxis)
        self.extend_figures(plot.get_figures())

    @property
    def coulomb_efficiency(self):
        data: RedoxFlowData = self.get_data()
        discharge = data.discharge_current_sec
        difference = data.charge_difference
        charge = data.charge_current_sec
        # efficiency_coulomb = (data.discharge_current_sec + data.charge_difference) / data.charge_current_sec * 100
        if charge > 0.0 and discharge > 0.0:
            efficiency_coulomb = 100 * 0.5 * ((difference * math.sqrt(4 * discharge * charge + difference ** 2)) /
                                              charge ** 2 + 2 * discharge / charge + (difference / charge) ** 2)
        elif charge > 0.0:
            efficiency_coulomb = 100 * difference / charge
            self.__log.warn('The storage was only charged. Thus, only a one-way efficiency could be calculated')
        elif discharge > 0.0:
            efficiency_coulomb = 100 * -difference / discharge
            self.__log.warn('The storage was only discharged. Thus, only a one-way efficiency could be calculated')
        else:
            efficiency_coulomb = numpy.nan
            self.__log.warn('Storage was neither charged nor discharged. Efficiency could not be calculated.')
        if not 0.0 < efficiency_coulomb < 100.0:
            self.__log.warn('Coulombic Efficiency should be between 0 % and 100 %, but is ' +
                            format_float(efficiency_coulomb) + ' %. Perhaps your simulation time range is too short '
                            'for calculating an accurate round trip efficiency.')
        return efficiency_coulomb



from simses.system.auxiliary.auxiliary import Auxiliary
from simses.system.power_electronics.dcdc_converter.abstract_dcdc_converter import DcDcConverter


class FixEfficiencyDcDcConverter(DcDcConverter):

    __volumetric_power_density = 143 * 10 ** 6  # W / m3
    __GRAVIMETRIC_POWER_DENSITY = 11000  # W/kg
    __SPECIFIC_SURFACE_AREA = 0.0001  # in m2 / W  # TODO add exact values

    # Exemplary value from:
    # (https://www.iisb.fraunhofer.de/en/research_areas/vehicle_electronics/dcdc_converters/High_Power_Density.html)

    def __init__(self, intermediate_circuit_voltage: float, efficiency: float = 0.98):
        super().__init__(intermediate_circuit_voltage)
        self.__dc_power_loss: float = 0
        self.__dc_power: float = 0
        self.__dc_current: float = 0
        self.__efficiency: float = efficiency

    def calculate_dc_current(self, dc_power_intermediate_circuit: float, storage_voltage: float) -> None:
        if self._is_charge(dc_power_intermediate_circuit):
            dc_power_storage = dc_power_intermediate_circuit * self.__efficiency
        else:
            dc_power_storage = dc_power_intermediate_circuit / self.__efficiency
        self.__dc_power = dc_power_storage
        self.__dc_power_loss = abs(dc_power_intermediate_circuit - dc_power_storage)
        dc_current_storage = dc_power_storage / storage_voltage
        self.__dc_current = dc_current_storage

    def reverse_calculate_dc_current(self, dc_power_storage: float, storage_voltage: float) -> None:
        if self._is_charge(dc_power_storage):
            dc_power_intermediate_circuit = dc_power_storage / self.__efficiency
        else:
            dc_power_intermediate_circuit = dc_power_storage * self.__efficiency
        self.__dc_power = dc_power_intermediate_circuit
        self.__dc_power_loss = abs(dc_power_storage - dc_power_intermediate_circuit)
        dc_current_intermediate_circuit = dc_power_intermediate_circuit / self.intermediate_circuit_voltage
        self.__dc_current = dc_current_intermediate_circuit

    @property
    def max_power(self) -> float:
        return 1e12

    @property
    def dc_power_loss(self):
        return self.__dc_power_loss

    @property
    def dc_power(self):
        return self.__dc_power

    @property
    def volume(self) -> float:
        # return self.max_power / self.__volumetric_power_density
        return 0

    @property
    def mass(self):
        return 0

    @property
    def surface_area(self) -> float:
        return 0

    @property
    def dc_current(self):
        return self.__dc_current

    def get_auxiliaries(self) -> [Auxiliary]:
        return list()

    def close(self) -> None:
        pass



from simses.commons.state.technology.fuel_cell import FuelCellState
from simses.technology.hydrogen.control.pressure.no_pressure_controller import NoPressureController
from simses.technology.hydrogen.fuel_cell.pressure.pressure_model import PressureModel


class NoPressureModel(PressureModel):

    def __init__(self):
        super().__init__()
        self.__pressure_cathode = 0  # barg
        self.__pressure_anode = 0  # barg
        self.__pressure_cathode_desire = 0  # barg
        self.__pressure_anode_desire = 0  # barg
        self.__hydrogen_inflow = 0  # mol/s
        self.__oxygen_inflow = 0  # mol/s
        self.__pressure_controller = NoPressureController()

    def calculate(self, time, state: FuelCellState) -> None:
        self.__pressure_anode = state.pressure_anode
        self.__pressure_cathode = state.pressure_cathode
        hydrogen_used = state.hydrogen_use
        oxygen_used = state.oxygen_use
        self.__hydrogen_inflow = self.__pressure_controller.calculate_n_h2_in(self.__pressure_anode, hydrogen_used)
        self.__oxygen_inflow = self.__pressure_controller.calculate_n_o2_in(self.__pressure_cathode, oxygen_used)

    def get_pressure_cathode_fc(self) -> float:
        return self.__pressure_cathode

    def get_pressure_anode_fc(self) -> float:
        return self.__pressure_anode

    def get_h2_inflow(self) -> float:
        return self.__hydrogen_inflow

    def get_o2_inflow(self) -> float:
        return self.__oxygen_inflow





import time as calctime
from simses.commons.profile.power.power_profile import PowerProfile
from simses.commons.config.simulation.energy_management import EnergyManagementConfig
from simses.commons.state.energy_management import EnergyManagementState
from simses.commons.state.system import SystemState
from simses.logic.energy_management.strategy.operation_priority import OperationPriority
from simses.logic.energy_management.strategy.operation_strategy import OperationStrategy


class SimplePeakShaving(OperationStrategy):
    """
    Basic Peak Shaving operation strategy: If the storage is almost full (> xy %),
    the storage is not charged anymore to avoid a misrepresent fulfillmentfactor
    """

    def __init__(self, power_profile: PowerProfile, ems_config: EnergyManagementConfig):
        super().__init__(OperationPriority.VERY_HIGH)
        self.__load_profile: PowerProfile = power_profile
        self.__max_power_monthly_mode: bool = ems_config.max_power_monthly_mode
        self.__max_power_yearly: float = ems_config.max_power
        self.__max_power_monthly: list = ems_config.max_power_monthly
        self.__max_power = 0
        self.__power: float = 0.0
        self.__soc_threshold: float = 0.9999
        self.__power_threshold: float = 0.0

    def next(self, time: float, system_state: SystemState, power_offset: float = 0) -> float:
        self.__max_power = self.__get_ps_limit(time)
        self.__power = self.__load_profile.next(time)
        net_power = self.__max_power - (self.__power + power_offset)
        if net_power < 0.0:
            return net_power
        elif system_state.soc < self.__soc_threshold and net_power > self.__power_threshold:
            return net_power
        else:
            return 0.0

    def __get_ps_limit(self, time: float) -> float:
        if self.__max_power_monthly_mode:
            month_index = int(calctime.strftime('%m', calctime.gmtime(time))) - 1
            max_power = 0
            for limit in self.__max_power_monthly:
                max_power += float(limit[month_index])
        else:
            max_power = self.__max_power_yearly
        return max_power

    def update(self, energy_management_state: EnergyManagementState) -> None:
        energy_management_state.load_power = self.__power
        energy_management_state.peakshaving_limit = self.__max_power

    def clear(self) -> None:
        self.__power = 0.0

    def close(self) -> None:
        self.__load_profile.close()



from simses.commons.cycle_detection.cycle_detector import CycleDetector
from simses.commons.log import Logger
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.cell.type import CellType
from simses.technology.lithium_ion.degradation.cyclic.cyclic_degradation import \
    CyclicDegradationModel


class SanyoNMCCyclicDegradationModel(CyclicDegradationModel):
    """Source: Schmalstieg, J., Käbitz, S., Ecker, M., & Sauer, D. U. (2014).
    A holistic aging model for Li (NiMnCo) O2 based 18650 lithium-ion batteries.
    Journal of Power Sources, 257, 325-334."""

    def __init__(self, cell_type: CellType, cycle_detector: CycleDetector):
        super().__init__(cell_type, cycle_detector)
        self.__log: Logger = Logger(type(self).__name__)
        self.__capacity_loss = 0
        self.__resistance_increase = 0
        self.__voltage_last_step_capacity = 3.6
        self.__voltage_last_step_resistance = 3.6
        self.__cycle_detector: CycleDetector = cycle_detector
        self.__ah_throughput_total_cap = 0

    def calculate_degradation(self, battery_state: LithiumIonState) -> None:
        # get doc, delta charge throughput and average voltage between cycles
        doc = self._cycle_detector.get_depth_of_cycle()
        qcell = battery_state.capacity / battery_state.nominal_voltage / self._cell.get_parallel_scale()
        delta_ah_throughput = qcell*doc
        voltage = self._cell.get_open_circuit_voltage(battery_state) / self._cell.get_serial_scale()
        v_average = (self.__voltage_last_step_capacity + voltage) / 2
        self.__voltage_last_step_capacity = voltage

        beta_cap = 7.348*10**(-3)*(v_average - 3.667)**2 + 7.6*10**(-4) + 4.081*10**(-3)*doc
        # implementation of discretization method as in MA Chuanqin Ni
        total_capacity_t1 = 1 - beta_cap * self.__ah_throughput_total_cap ** 0.5  # in p.u.
        total_capacity_t2 = 1 - beta_cap * (self.__ah_throughput_total_cap + delta_ah_throughput) ** 0.5  # in p.u.
        capacity_loss = total_capacity_t1 - total_capacity_t2

        self.__ah_throughput_total_cap += delta_ah_throughput
        self.__capacity_loss = capacity_loss * self._cell.get_nominal_capacity()  # Ah

    def calculate_resistance_increase(self, battery_state: LithiumIonState) -> None:
        # get doc, delta charge throughput and average voltage between cycles
        doc = self._cycle_detector.get_depth_of_cycle()
        qcell = battery_state.capacity / battery_state.nominal_voltage / self._cell.get_parallel_scale()
        delta_ah_throughput = qcell * doc
        voltage = self._cell.get_open_circuit_voltage(battery_state) / self._cell.get_serial_scale()
        v_average = (self.__voltage_last_step_resistance + voltage) / 2
        self.__voltage_last_step_resistance = voltage

        beta_res = 2.153*10**(-4)*(v_average - 3.725)**2 - 1.521*10**(-5) + 2.798*10**(-4)*doc
        # implementation of discretization method as in MA Chuanqin Ni
        resistance_increase = beta_res * delta_ah_throughput # in p.u.

        self.__resistance_increase = resistance_increase  # pu

    def get_degradation(self) -> float:
        capacity_loss = self.__capacity_loss
        self.__capacity_loss = 0    # Set value to 0, because cyclic losses are not calculated in each step
        return capacity_loss

    def get_resistance_increase(self) -> float:
        resistance_increase = self.__resistance_increase
        self.__resistance_increase = 0 # Set value to 0, because cyclic losses are not calculated in each step
        return resistance_increase

    def reset(self, lithium_ion_state: LithiumIonState) -> None:
        self.__capacity_loss = 0
        self.__resistance_increase = 0

    def close(self) -> None:
        self.__log.close()



class TimeValue:
    """
    TimeValue is a data class with a timestamp connected to value.
    """

    def __init__(self, tstmp: float, value: float):
        """
        Constructor of TimeValue

        Parameters
        ----------
        tstmp :
            epoch timestamp in s
        value :
            value connected to timestamp
        """
        self.__tstmp: float = tstmp
        self.__value: float = value

    @property
    def time(self) -> float:
        return self.__tstmp

    @time.setter
    def time(self, value: float) -> None:
        self.__tstmp = value

    @property
    def value(self) -> float:
        return self.__value

    @value.setter
    def value(self, value: float) -> None:
        self.__value = value

    def __str__(self) -> str:
        return '\n(' + str(self.time) + ', ' + str(self.value) + ')'

    def __repr__(self):
        return self.__str__()

    @staticmethod
    def sort_by_time(data: list, descending: bool = False) -> None:
        """
        In-place sorting of list with TimeValue objects by time (ascending)

        Parameters
        ----------
        descending :
            reverse sorting of data list, default: False
        data :
            list of TimeValue objects

        Returns
        -------

        """
        data.sort(key=lambda x: x.time, reverse=descending)



import numpy as np

from simses.commons.utils.utilities import format_float


class Description:

    class Technical:
        ROUND_TRIP_EFFICIENCY: str = 'Efficiency round trip'
        MEAN_SOC: str = 'SOC mean'
        MAX_SOC: str = 'SOC max'
        MIN_SOC: str = 'SOC min'
        NUMBER_CHANGES_SIGNS: str = 'Number of changes of signs per day'
        RESTING_TIME_AVG: str = 'Avg. length of resting times'
        ENERGY_CHANGES_SIGN: str = 'Pos. energy between changes of sign'
        FULFILLMENT_AVG: str = 'Avg. Fulfillment Factor'
        EQUIVALENT_FULL_CYCLES: str = 'Equivalent full cycles'
        DEPTH_OF_DISCHARGE: str = 'Avg. depth of cycle for discharge'
        REMAINING_CAPACITY: str = 'Remaining capacity'
        SELF_CONSUMPTION_RATE: str = 'Self-consumption rate'
        SELF_SUFFICIENCY: str = 'Self-sufficiency'
        ENERGY_THROUGHPUT: str = 'Energy throughput'
        C_RATE_CHARGING: str = 'Avg. C-rate charging'
        C_RATE_DISCHARGING: str = 'Avg. C-rate discharging'
        COULOMB_EFFICIENCY: str = 'Coulomb efficiency'
        MAX_GRID_POWER: str = 'Max. grid power'
        POWER_ABOVE_PEAK_MAX: str = 'Max. grid power above peak'
        POWER_ABOVE_PEAK_AVG: str = 'Average power above peak'
        ENERGY_ABOVE_PEAK_MAX: str = 'Max. energy event above peak'
        ENERGY_ABOVE_PEAK_AVG: str = 'Average energy event above peak'
        NUMBER_ENERGY_EVENTS: str = 'Number of energy events above peak'
        MAX_LOAD_DC_POWER_ADDITIONAL: str = 'Max. load of additional dc power'
        MAX_GENERATION_DC_POWER_ADDITIONAL: str = 'Max. generation of additional dc power'
        ACDC_EFFICIENCY_DISCHARGE: str = 'AC_DC efficiency discharge'
        ACDC_EFFICIENCY_CHARGE: str = 'AC_DC efficiency charge'
        ACDC_EFFICIENCY: str = 'AC_DC efficiency total'
        DCDC_EFFICIENCY_DISCHARGE: str = 'DC_DC efficiency discharge'
        DCDC_EFFICIENCY_CHARGE: str = 'DC_DC efficiency charge'
        DCDC_EFFICIENCY: str = 'DC_DC efficiency total'
        SYSTEM_TEMPORAL_UTILIZATION: str = 'System temporal utilization in p.u.'
        E_RATE_CHARGING: str = 'Avg. E-rate charging'
        E_RATE_DISCHARGING: str = 'Avg. E-rate discharging'
        STORAGE_TECHNOLOGY_TEMPORAL_UTILIZATION: str = 'Storage technology temporal utilization in p.u.'
        PE_EFFICIENCY: str = 'Power electronics efficiency total'
        H2_PRODUCTION_EFFICIENCY_LHV: str = 'Hydrogen production efficiency relative to LHV'
        TOTAL_H2_PRODUCTION_KG: str = 'Total mass of hydrogen'
        TOTAL_H2_PRODUCTION_NM: str = 'Total volume hydrogen'
        ENERGY_H2_LHV: str = 'Energy of Hydrogen relative to LHV'
        ENERGY_H2_DRYING: str = 'Energy for Drying of Hydrogen'
        ENERGY_H2_COMPRESSION: str = 'Energy for compression of hydrogen'
        ENERGY_H2_REACTION: str = 'Energy for chemical reaction'
        ENERGY_WATER_HEATING: str = 'Energy for heating of water'
        ENERGY_WATER_CIRCULATION: str = 'Energy for water circulation'
        SOH: str = 'Sate of Health'
        ENERGY_IDM_BOUGHT: str = 'Energy bought on intraday market'
        ENERGY_IDM_SOLD: str = 'Energy sold on intraday market'

    class Economical:
        CASHFLOW: str = 'Cashflow each year'
        NET_PRESENT_VALUE: str = 'NPV'
        INTERNAL_RATE_OF_RETURN: str = 'IRR'
        PROFITABILITY_INDEX: str = 'Profitability Index'
        RETURN_ON_INVEST: str = 'ROI'
        LEVELIZED_COST_OF_STORAGE: str = 'Levelized Cost of Storage'
        INVESTMENT_COSTS: str = 'Investment Costs'
        DISCOUNT_RATE: str = 'Discount Rate'

        class FCR:
            PRICE_AVERAGE: str = 'Average FCR Price Each Year'
            REVENUE_YEARLY: str = 'FCR Revenue Each Year'
            POWER_BID_AVERAGE: str = 'Average FCR Power Bid Each Year'

        class Intraday:
            PRICE_AVERAGE: str = 'Average Intraday Revenue Price Each Year'
            REVENUE_YEARLY: str = 'Intraday Revenue Each Year'
            POWER_AVERAGE: str = 'Average Intraday Power Each Year'

        class SCI:
            COST_WITHOUT_STORAGE: str = 'Electricity cost each year without storage system'
            COST_WITH_STORAGE: str = 'Electricity cost each year with storage system'
            COST_ELECTRICITY: str = 'Electricity Costs'
            PV_FEED_IN_TARIFF: str = 'PV Feed In Tariff'

        class DemandCharges:
            COST_WITHOUT_STORAGE: str = 'Demand charges each year without storage system'
            COST_WITH_STORAGE: str = 'Demand charges each year with storage system'
            CYCLE: str = 'Demand Charge Billing Cycle'
            PRICE: str = 'Demand Charge Price'
            INTERVAL: str = 'Demand Charge Average Interval'

        class OperationAndMaintenance:
            ANNUAl_O_AND_M_COST: str = 'Annual Op. and Maint. Cost '
            O_AND_M_COST: str = 'Op. and Maint. Cost each year'

        class ElectricityConsumption:
            ELECTRICITY_COST_RENEWABLE: str = 'Electricity cost renewable'
            ELECTRICITY_COST_GRID: str = 'Electricity cost grid'
            TOTAL_ELECTRICITY_COST: str = 'Total electricity Cost'
            ELECTRICITY_PRICE: str = 'Electricity price'


class Unit:
    NONE: str = ''
    PERCENTAGE: str = '%'
    MINUTES: str = 'min'
    EURO: str = 'EUR'
    WATT: str = 'W'
    KILOWATT: str = 'kW'
    KWH: str = 'kWh'
    EURO_PER_KW: str = 'EUR / kW'
    EURO_PER_MWH: str = 'EUR / MWh'
    EURO_PER_KWH: str = 'EUR / kWh'
    EURO_PER_KW_DAY: str = 'EUR / kW / d'
    NCM : str = 'Nm^3'
    KG: str = 'kg'
    ONE_PER_HOUR: str = '1/h'


class EvaluationResult:
    """Provides a structure for evaluation results in order to organize data management for printing to
    the console, exporting to csv files, etc.."""

    def __init__(self, description: str, unit: str, value):
        self.__description: str = description
        self.__unit: str = unit
        self.__value = value

    @property
    def description(self) -> str:
        """Description of the result"""
        return self.__description

    @property
    def unit(self) -> str:
        """Unit of the result"""
        return self.__unit

    @property
    def value(self) -> str:
        """Value of the result"""
        if isinstance(self.__value, (int, float, complex)) and not np.isnan(self.__value):
            return format_float(self.__value, decimals=2)
        elif isinstance(self.__value, (list, np.ndarray)):
            return str([round(value, 2) for value in self.__value])
        else:
            return str(self.__value)

    def to_csv(self) -> [str]:
        """Returns EvaluationResult as a list of strings."""
        res = list()
        res.append(self.description)
        res.append(self.value)
        res.append(self.unit)
        return res

    def to_console(self) -> str:
        """Returns EvaluationResult as a string in a format that is suitable for printing to the console."""
        res = ''
        res += '{:<50}'.format(self.description)
        res += self.value + ' '
        res += self.unit
        return res

    @classmethod
    def get_header(cls) -> [str]:
        """Returns the header of EvaluationResult as a list of strings."""
        return [str(Description.__name__), 'Value', str(Unit.__name__)]

    def __str__(self):
        return self.to_csv()

    def __repr__(self):
        return self.to_csv()



import pandas

from simses.analysis.data.abstract_data import Data
from simses.commons.state.technology.fuel_cell import FuelCellState
from simses.commons.config.simulation.general import GeneralSimulationConfig


class FuelCellData(Data):
    """
    Provides time series data from FuelCellState
    """

    def __init__(self, config: GeneralSimulationConfig, data: pandas.DataFrame):
        super().__init__(config, data)
        self.__config = config

    @property
    def id(self) -> str:
        return str(int(self._get_first_value(FuelCellState.SYSTEM_AC_ID))) + '.' + \
               str(int(self._get_first_value(FuelCellState.SYSTEM_DC_ID)))

    @property
    def time(self):
        return self._get_data(FuelCellState.TIME)

    @property
    def power(self):
        return self.current * self.voltage

    @property
    def dc_power(self):
        return self.power

    @property
    def energy_difference(self):
        pass

    @property
    def soc(self):
        pass

    @property
    def capacity(self):
        pass

    @property
    def state_of_health(self):
        pass

    @property
    def storage_fulfillment(self):
        return self._get_data(FuelCellState.FULFILLMENT)

    @property
    def current(self):
        return self._get_data(FuelCellState.CURRENT)

    @property
    def voltage(self):
        return self._get_data(FuelCellState.VOLTAGE)

    @property
    def current_density(self):
        return self._get_data(FuelCellState.CURRENT_DENSITY)

    @property
    def hydrogen_use(self):
        return self._get_data(FuelCellState.HYDROGEN_USE)

    @property
    def pressure_anode(self):
        return self._get_data(FuelCellState.PRESSURE_ANODE)

    @property
    def pressure_cathode(self):
        return self._get_data(FuelCellState.PRESSURE_CATHODE)

    @property
    def temperature(self):
        return self._get_data(FuelCellState.TEMPERATURE)

    @classmethod
    def get_system_data(cls, path: str, config: GeneralSimulationConfig) -> list:
        system_data: [pandas.DataFrame] = cls._get_system_data_for(path, FuelCellState, FuelCellState.TIME,
                                                                   FuelCellState.SYSTEM_AC_ID,
                                                                   FuelCellState.SYSTEM_DC_ID)
        res: [FuelCellData] = list()
        for data in system_data:
            res.append(FuelCellData(config, data))
        return res



from simses.commons.profile.file import FileProfile
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.config.simulation.profile import ProfileConfig
from simses.commons.profile.technical.technical import TechnicalProfile
from copy import copy
import numpy as np

class BinaryProfile(TechnicalProfile):
    """
    BinaryProfile is a subclass of TechnicalProfile (like SocProfile). The binary profile should contain a timestep and
    a binary profile (0 or 1). This class is used for temporary unavailable storage systems, e.g. mobile applications
    like electric vehicles. Here zeros indicate that the car is on the road respectively not "at home" and ones indicate
    that the car is "at home" and could be charged.
    """

    def __init__(self, config: GeneralSimulationConfig, profile_config: ProfileConfig):
        super().__init__()
        self.__file: FileProfile = FileProfile(config, profile_config.binary_profile_file, delimiter=',')
        self.__file_copy = None


    def next(self, time: float) -> float:
        return self.__file.next(time)


    def next_change_in_binary(self, time: float, current_binary: bool) -> float:

        self.__file_copy = copy(self.__file) # Copy of file because the last value is saved in next-function
        time_copy = copy(time)
        current_value = current_binary
        while bool(np.ceil(current_value)) == current_binary and time_copy <= self.__file_copy._FileProfile__end:
            # maximum until end of file
            time_copy += 1
            current_value = self.__file_copy.next(time_copy)

        self.__file_copy = None

        return time_copy


    def close(self):
        self.__file.close()



from configparser import ConfigParser
from datetime import datetime

import numpy as np
import pandas
import pytest

from simses.analysis.data.energy_management import EnergyManagementData
from simses.analysis.data.system import SystemData
from simses.analysis.evaluation.economic.revenue_stream.fcr_revenue_stream import FCRRevenue
from simses.commons.config.analysis.economic import EconomicAnalysisConfig
from simses.commons.config.simulation.general import GeneralSimulationConfig
# fixed input parameters
from simses.commons.state.energy_management import EnergyManagementState
from simses.commons.state.system import SystemState

fcr_price = 0.3  # EUR / (kW * day)
years = 5
timestep = 3600
unix_timestamp_start_2020 = 1577836800
n = int(years * 365 * 24 * 60 * 60 / timestep)
unix_timestamp_start_end = unix_timestamp_start_2020 + n*timestep


def create_general_config() -> GeneralSimulationConfig:
    simulation_config: ConfigParser = ConfigParser()
    simulation_config.add_section('GENERAL')
    simulation_config.set('GENERAL', 'TIME_STEP', str(timestep))
    simulation_config.set('GENERAL', 'START',
                          datetime.fromtimestamp(unix_timestamp_start_2020).strftime('%Y-%m-%d %H:%M:%S'))
    simulation_config.set('GENERAL', 'END',
                          datetime.fromtimestamp(unix_timestamp_start_end).strftime('%Y-%m-%d %H:%M:%S'))
    return GeneralSimulationConfig(simulation_config)


def create_economic_analysis_config() -> EconomicAnalysisConfig:
    analysis_config: ConfigParser = ConfigParser()
    analysis_config.add_section('ECONOMIC_ANALYSIS')
    analysis_config.set('ECONOMIC_ANALYSIS', 'FCR_PRICE', str(fcr_price))
    analysis_config.set('ECONOMIC_ANALYSIS', 'FCR_USE_PRICE_TIMESERIES', 'False')
    return EconomicAnalysisConfig(analysis_config)


@pytest.mark.parametrize('fcr_power_const', [-1e6, 1e6])
def test_fcr_revenue_stream(fcr_power_const):
    """Performs a unit test by comparing the expected result for a generic
    time series with the actual result."""

    # set up test data
    time_test = pandas.array([unix_timestamp_start_2020 + timestep * i for i in range(n)])
    fcr_power = pandas.array([fcr_power_const] * n)
    battery_power = pandas.array([0] * n)

    # calculate expected result:
    expected_result = np.array([abs(fcr_power_const) * 365 * fcr_price / 1000] * years)

    # build configs for testing
    gen_sim_config = create_general_config()
    economic_config = create_economic_analysis_config()
    market_profile_config = None

    # create data for testing
    energy_management_dict = {EnergyManagementState.TIME: time_test, EnergyManagementState.FCR_MAX_POWER: fcr_power}
    energy_management_data = EnergyManagementData(gen_sim_config, pandas.DataFrame(energy_management_dict))
    system_dict = {SystemState.TIME: time_test, SystemState.AC_POWER_DELIVERED: battery_power}
    system_data = SystemData(gen_sim_config, pandas.DataFrame(system_dict))

    fcr_revenue_stream = FCRRevenue(energy_management_data, system_data, economic_config, gen_sim_config, market_profile_config)
    result1 = fcr_revenue_stream.get_cashflow()
    assert list(result1.round(0)) == list(expected_result.round(0))


from simses.commons.constants import Hydrogen
from simses.commons.state.technology.electrolyzer import ElectrolyzerState
from simses.technology.hydrogen.electrolyzer.stack.pem_analytic.membrane_model import \
    PemMembraneModel


class PemFluidModel:

    def __init__(self, membrane_model: PemMembraneModel):
        self.__membrane_model: PemMembraneModel = membrane_model

    def get_h2_net_cathode(self, state: ElectrolyzerState, current_cell: float) -> float:
        return self.__h2_generation_cell(current_cell) - \
               self.__membrane_model.get_h2_permeation_for_cell(state, current_cell) - \
               2 * self.__membrane_model.get_o2_permeation_for_cell(state, current_cell)  # mol/s

    def get_o2_net_anode(self, state: ElectrolyzerState, current_cell: float) -> float:
        return self.__o2_generation_cell(current_cell) - \
               self.__membrane_model.get_o2_permeation_for_cell(state, current_cell / self.__membrane_model.get_geometric_area_cell())  # mol/s

    def get_h2o_net_use_cell(self, state: ElectrolyzerState, current_cell: float) -> float:
        return - (self.__h2o_net_anode(current_cell) +
                  self.__h2o_net_cathode(state, current_cell))  # mol/s

    def __h2_generation_cell(self, current_cell) -> float:
        return current_cell / (2 * Hydrogen.FARADAY_CONST)  # mol/s

    def __o2_generation_cell(self, current_cell) -> float:
        return current_cell / (4 * Hydrogen.FARADAY_CONST)  # mol/s

    def __h2o_use_cell(self, current_cell) -> float:
        return current_cell / (2 * Hydrogen.FARADAY_CONST)  # mol/s

    def __h2o_net_cathode(self, state: ElectrolyzerState, current_cell: float) -> float:
        return self.__membrane_model.get_h2o_permeation_for_cell(current_cell) + \
               2 * self.__membrane_model.get_o2_permeation_for_cell(state, current_cell)  # mol/s

    def __h2o_net_anode(self, current_cell) -> float:
        return - self.__h2o_use_cell(current_cell) - \
               self.__membrane_model.get_h2o_permeation_for_cell(current_cell)  # mol/s

    def get_h2_generation_cell(self, current_cell) -> float:
        return self.__h2_generation_cell(current_cell)

    def get_o2_generation_cell(self, current_cell) -> float:
        return self.__o2_generation_cell(current_cell)

    def get_h2o_use_cell(self, current_cell) -> float:
        return self.__h2o_use_cell(current_cell)



from configparser import ConfigParser

from simses.commons.config.generation.analysis import AnalysisConfigGenerator
from simses.commons.config.generation.simulation import SimulationConfigGenerator
from simses.simulation.batch_processing import BatchProcessing


class ExampleBatchProcessing(BatchProcessing):

    """
    This is just a simple example on how to use BatchProcessing.
    """

    def __init__(self):
        super().__init__(do_simulation=True, do_analysis=True)

    def _setup_config(self) -> dict:
        # Example for config setup
        config_generator: SimulationConfigGenerator = SimulationConfigGenerator()
        # example: loading default config as base (not necessary)
        config_generator.load_default_config()
        # defining parameters
        capacity: float = 5000.0
        ac_power: float = 5000.0
        voltage_ic: float = 600.0
        # generating config options
        storage_1: str = config_generator.add_generic_cell(capacity=capacity)
        storage_2: str = config_generator.add_lithium_ion_battery(capacity=capacity, cell_type='SonyLFP')
        storage_3: str = config_generator.add_lithium_ion_battery(capacity=capacity, cell_type='PanasonicNCA')
        storage_4: str = config_generator.add_lithium_ion_battery(capacity=capacity, cell_type='MolicelNMC')
        dcdc_1: str = config_generator.add_fix_efficiency_dcdc(efficiency=0.98)
        acdc_1: str = config_generator.add_fix_efficiency_acdc()
        housing_1: str = config_generator.add_no_housing()
        hvac_1: str = config_generator.add_no_hvac()
        # generating storage systems
        config_generator.clear_storage_system_ac()
        ac_system_1: str = config_generator.add_storage_system_ac(ac_power, voltage_ic, acdc_1, housing_1, hvac_1)
        config_generator.clear_storage_system_dc()
        config_generator.add_storage_system_dc(ac_system_1, dcdc_1, storage_1)
        # printing config (if wanted)
        config_generator.show()
        # get config ready to be passed to SimSES
        config: ConfigParser = config_generator.get_config()
        # setting up multiple configurations with manual naming of simulations
        config_set: dict = dict()
        config_generator.clear_storage_system_dc()
        config_generator.add_storage_system_dc(ac_system_1, dcdc_1, storage_1)
        config_set['storage_1'] = config_generator.get_config()
        config_generator.clear_storage_system_dc()
        config_generator.add_storage_system_dc(ac_system_1, dcdc_1, storage_2)
        config_set['storage_2'] = config_generator.get_config()
        config_generator.clear_storage_system_dc()
        config_generator.add_storage_system_dc(ac_system_1, dcdc_1, storage_3)
        config_set['storage_3'] = config_generator.get_config()
        config_generator.clear_storage_system_dc()
        config_generator.add_storage_system_dc(ac_system_1, dcdc_1, storage_4)
        config_set['storage_4'] = config_generator.get_config()
        config_generator.clear_storage_system_dc()
        config_generator.add_storage_system_dc(ac_system_1, dcdc_1, storage_1)
        config_generator.add_storage_system_dc(ac_system_1, dcdc_1, storage_2)
        config_set['hybrid_1'] = config_generator.get_config()
        config_generator.clear_storage_system_dc()
        config_generator.add_storage_system_dc(ac_system_1, dcdc_1, storage_3)
        config_generator.add_storage_system_dc(ac_system_1, dcdc_1, storage_4)
        config_set['hybrid_2'] = config_generator.get_config()
        return config_set

    def _analysis_config(self) -> ConfigParser:
        config_generator: AnalysisConfigGenerator = AnalysisConfigGenerator()
        config_generator.print_results(False)
        config_generator.do_plotting(True)
        return config_generator.get_config()

    def clean_up(self) -> None:
        pass


if __name__ == "__main__":
    batch_processing: BatchProcessing = ExampleBatchProcessing()
    batch_processing.run()
    batch_processing.clean_up()
